# -*- coding: utf-8 -*-
"""
A module containing NodeEditor's class for representing `Node`.
"""
from NodeEditor.nodeeditor.node_graphics_node import QDMGraphicsNode
from NodeEditor.nodeeditor.node_content_widget import QDMNodeContentWidget
from NodeEditor.nodeeditor.node_socket import *
from NodeEditor.nodeeditor.utils import dumpException, pp

DEBUG = False


class Node(Serializable):
    """
    Class representing `Node` in the `Scene`.
    """
    GraphicsNode_class = QDMGraphicsNode
    NodeContent_class = QDMNodeContentWidget
    Socket_class = Socket

    def __init__(self, scene:'Scene', title:str="Undefined Node", inputs:list=[], outputs:list=[]):
        """

        :param scene: reference to the :class:`~nodeeditor.node_scene.Scene`
        :type scene: :class:`~nodeeditor.node_scene.Scene`
        :param title: Node Title shown in Scene
        :type title: str
        :param inputs: list of :class:`~nodeeditor.node_socket.Socket` types from which the `Sockets` will be auto created
        :param outputs: list of :class:`~nodeeditor.node_socket.Socket` types from which the `Sockets` will be auto created

        :Instance Attributes:

            - **scene** - reference to the :class:`~nodeeditor.node_scene.Scene`
            - **grNode** - Instance of :class:`~nodeeditor.node_graphics_node.QDMGraphicsNode` handling graphical representation in the ``QGraphicsScene``. Automatically created in constructor
            - **content** - Instance of :class:`~nodeeditor.node_graphics_content.QDMGraphicsContent` which is child of ``QWidget`` representing container for all inner widgets inside of the Node. Automatically created in constructor
            - **inputs** - list containin Input :class:`~nodeeditor.node_socket.Socket` instances
            - **outputs** - list containin Output :class:`~nodeeditor.node_socket.Socket` instances

        """
        super().__init__()
        self._title = title
        self.scene = scene

        # just to be sure, init these variables
        self.content = None
        self.grNode = None

        self.initInnerClasses()
        self.initSettings()

        self.title = title

        self.scene.addNode(self)
        self.scene.grScene.addItem(self.grNode)

        # create socket for inputs and outputs
        self.inputs = []
        self.outputs = []
        self.initSockets(inputs, outputs)

        # dirty and evaluation
        self._is_dirty = False
        self._is_invalid = False

    def __str__(self):
        return "<%s:%s %s..%s>" % (self.title, self.__class__.__name__,hex(id(self))[2:5], hex(id(self))[-3:])

    @property
    def title(self):
        """
        Title shown in the scene

        :getter: return current Node title
        :setter: sets Node title and passes it to Graphics Node class
        :type: ``str``
        """
        return self._title

    @title.setter
    def title(self, value):
        self._title = value
        self.grNode.title = self._title

    @property
    def pos(self):
        """
        Retrieve Node's position in the Scene

        :return: Node position
        :rtype: ``QPointF``
        """
        return self.grNode.pos()        # QPointF

    def setPos(self, x:float, y:float):
        """
        Sets position of the Graphics Node

        :param x: X `Scene` position
        :param y: Y `Scene` position
        """
        self.grNode.setPos(x, y)


    def initInnerClasses(self):
        """Sets up graphics Node (PyQt) and Content Widget"""
        node_content_class = self.getNodeContentClass()
        graphics_node_class = self.getGraphicsNodeClass()
        if node_content_class is not None: self.content = node_content_class(self)
        if graphics_node_class is not None: self.grNode = graphics_node_class(self)

    def getNodeContentClass(self):
        """Returns class representing nodeeditor content"""
        return self.__class__.NodeContent_class

    def getGraphicsNodeClass(self):
        return self.__class__.GraphicsNode_class

    def initSettings(self):
        """Initialize properties and socket information"""
        self.socket_spacing = 22

        self.input_socket_position = LEFT_BOTTOM
        self.output_socket_position = RIGHT_TOP
        self.input_multi_edged = False
        self.output_multi_edged = True
        self.socket_offsets = {
            LEFT_BOTTOM: -1,
            LEFT_CENTER: -1,
            LEFT_TOP: -1,
            RIGHT_BOTTOM: 1,
            RIGHT_CENTER: 1,
            RIGHT_TOP: 1,
        }

    def initSockets(self, inputs:list, outputs:list, reset:bool=True):
        """
        Create sockets for inputs and outputs

        :param inputs: list of Socket Types (int)
        :type inputs: ``list``
        :param outputs: list of Socket Types (int)
        :type outputs: ``list``
        :param reset: if ``True`` destroys and removes old `Sockets`
        :type reset: ``bool``
        """

        if reset:
            # clear old sockets
            if hasattr(self, 'inputs') and hasattr(self, 'outputs'):
                # remove grSockets from scene
                for socket in (self.inputs+self.outputs):
                    self.scene.grScene.removeItem(socket.grSocket)
                self.inputs = []
                self.outputs = []

        # create new sockets
        counter = 0
        for item in inputs:
            socket = self.__class__.Socket_class(
                node=self, index=counter, position=self.input_socket_position,
                socket_type=item, multi_edges=self.input_multi_edged,
                count_on_this_node_side=len(inputs), is_input=True
            )
            counter += 1
            self.inputs.append(socket)

        counter = 0
        for item in outputs:
            socket = self.__class__.Socket_class(
                node=self, index=counter, position=self.output_socket_position,
                socket_type=item, multi_edges=self.output_multi_edged,
                count_on_this_node_side=len(outputs), is_input=False
            )
            counter += 1
            self.outputs.append(socket)


    def onEdgeConnectionChanged(self, new_edge:'Edge'):
        """
        Event handling that any connection (`Edge`) has changed. Currently not used...

        :param new_edge: reference to the changed :class:`~nodeeditor.node_edge.Edge`
        :type new_edge: :class:`~nodeeditor.node_edge.Edge`
        """
        pass

    def onInputChanged(self, socket:'Socket'):
        """Event handling when Node's input Edge has changed. We auto-mark this `Node` to be `Dirty` with all it's
        descendants

        :param socket: reference to the changed :class:`~nodeeditor.node_socket.Socket`
        :type socket: :class:`~nodeeditor.node_socket.Socket`
        """
        self.markDirty()
        self.markDescendantsDirty()

    def onDeserialized(self, data:dict):
        """Event manually called when this node was deserialized. Currently called when node is deserialized from scene
        Passing `data` containing the data which have been deserialized """
        pass

    def onDoubleClicked(self, event):
        """Event handling double click on Graphics Node in `Scene`"""
        pass

    def doSelect(self, new_state:bool=True):
        """Shortcut method for selecting/deselecting the `Node`

        :param new_state: ``True`` if you want to select the `Node`. ``False`` if you want to deselect the `Node`
        :type new_state: ``bool``
        """
        self.grNode.doSelect(new_state)

    def isSelected(self):
        """Returns ``True`` if current `Node` is selected"""
        return self.grNode.isSelected()

    def getSocketPosition(self, index:int, position:int, num_out_of:int=1) -> '(x, y)':
        """
        Get the relative `x, y` position of a :class:`~nodeeditor.node_socket.Socket`. This is used for placing
        the `Graphics Sockets` on `Graphics Node`.

        :param index: Order number of the Socket. (0, 1, 2, ...)
        :type index: ``int``
        :param position: `Socket Position Constant` describing where the Socket is located. See :ref:`socket-position-constants`
        :type position: ``int``
        :param num_out_of: Total number of Sockets on this `Socket Position`
        :type num_out_of: ``int``
        :return: Position of described Socket on the `Node`
        :rtype: ``x, y``
        """
        x = self.socket_offsets[position] if (position in (LEFT_TOP, LEFT_CENTER, LEFT_BOTTOM)) else self.grNode.width + self.socket_offsets[position]

        if position in (LEFT_BOTTOM, RIGHT_BOTTOM):
            # start from bottom
            y = self.grNode.height - self.grNode.edge_roundness - self.grNode.title_vertical_padding - index * self.socket_spacing
        elif position in (LEFT_CENTER, RIGHT_CENTER):
            num_sockets = num_out_of
            node_height = self.grNode.height
            top_offset = self.grNode.title_height + 2 * self.grNode.title_vertical_padding + self.grNode.edge_padding
            available_height = node_height - top_offset

            total_height_of_all_sockets = num_sockets * self.socket_spacing
            new_top = available_height - total_height_of_all_sockets

            # y = top_offset + index * self.socket_spacing + new_top / 2
            y = top_offset + available_height/2.0 + (index-0.5)*self.socket_spacing
            if num_sockets > 1:
                y -= self.socket_spacing * (num_sockets-1)/2

        elif position in (LEFT_TOP, RIGHT_TOP):
            # start from top
            y = self.grNode.title_height + self.grNode.title_vertical_padding + self.grNode.edge_roundness + index * self.socket_spacing
        else:
            # this should never happen
            y = 0

        return [x, y]

    def getSocketScenePosition(self, socket:'Socket') -> '(x, y)':
        """
        Get absolute Socket position in the Scene

        :param socket: `Socket` which position we want to know
        :return: (x, y) Socket's scene position
        """
        nodepos = self.grNode.pos()
        socketpos = self.getSocketPosition(socket.index, socket.position, socket.count_on_this_node_side)
        return (nodepos.x() + socketpos[0], nodepos.y() + socketpos[1])

    def updateConnectedEdges(self):
        """Recalculate (Refresh) positions of all connected `Edges`. Used for updating Graphics Edges"""
        for socket in self.inputs + self.outputs:
            # if socket.hasEdge():
            for edge in socket.edges:
                edge.updatePositions()

    def remove(self):
        """
        Safely remove this Node
        """
        if DEBUG: print("> Removing Node", self)
        if DEBUG: print(" - remove all edges from sockets")
        for socket in (self.inputs+self.outputs):
            # if socket.hasEdge():
            for edge in socket.edges.copy():
                if DEBUG: print("    - removing from socket:", socket, "edge:", edge)
                edge.remove()
        if DEBUG: print(" - remove grNode")
        self.scene.grScene.removeItem(self.grNode)
        self.grNode = None
        if DEBUG: print(" - remove node from the scene")
        self.scene.removeNode(self)
        if DEBUG: print(" - everything was done.")


    # node evaluation stuff

    def isDirty(self) -> bool:
        """Is this node marked as `Dirty`

        :return: ``True`` if `Node` is marked as `Dirty`
        :rtype: ``bool``
        """
        return self._is_dirty

    def markDirty(self, new_value:bool=True):
        """Mark this `Node` as `Dirty`. See :ref:`evaluation` for more

        :param new_value: ``True`` if this `Node` should be `Dirty`. ``False`` if you want to un-dirty this `Node`
        :type new_value: ``bool``
        """
        self._is_dirty = new_value
        if self._is_dirty: self.onMarkedDirty()

    def onMarkedDirty(self):
        """Called when this `Node` has been marked as `Dirty`. This method is supposed to be overriden"""
        pass

    def markChildrenDirty(self, new_value:bool=True):
        """Mark all first level children of this `Node` to be `Dirty`. Not this `Node` it self. Not other descendants

        :param new_value: ``True`` if children should be `Dirty`. ``False`` if you want to un-dirty children
        :type new_value: ``bool``
        """
        for other_node in self.getChildrenNodes():
            other_node.markDirty(new_value)

    def markDescendantsDirty(self, new_value:bool=True):
        """Mark all children and descendants of this `Node` to be `Dirty`. Not this `Node` it self

        :param new_value: ``True`` if children and descendants should be `Dirty`. ``False`` if you want to un-dirty children and descendants
        :type new_value: ``bool``
        """
        for other_node in self.getChildrenNodes():
            other_node.markDirty(new_value)
            other_node.markDescendantsDirty(new_value)

    def isInvalid(self) -> bool:
        """Is this node marked as `Invalid`?

        :return: ``True`` if `Node` is marked as `Invalid`
        :rtype: ``bool``
        """
        return self._is_invalid

    def markInvalid(self, new_value:bool=True):
        """Mark this `Node` as `Invalid`. See :ref:`evaluation` for more

        :param new_value: ``True`` if this `Node` should be `Invalid`. ``False`` if you want to make this `Node` valid
        :type new_value: ``bool``
        """
        self._is_invalid = new_value
        if self._is_invalid: self.onMarkedInvalid()

    def onMarkedInvalid(self):
        """Called when this `Node` has been marked as `Invalid`. This method is supposed to be overriden"""
        pass

    def markChildrenInvalid(self, new_value:bool=True):
        """Mark all first level children of this `Node` to be `Invalid`. Not this `Node` it self. Not other descendants

        :param new_value: ``True`` if children should be `Invalid`. ``False`` if you want to make children valid
        :type new_value: ``bool``
        """
        for other_node in self.getChildrenNodes():
            other_node.markInvalid(new_value)

    def markDescendantsInvalid(self, new_value:bool=True):
        """Mark all children and descendants of this `Node` to be `Invalid`. Not this `Node` it self

        :param new_value: ``True`` if children and descendants should be `Invalid`. ``False`` if you want to make children and descendants valid
        :type new_value: ``bool``
        """
        for other_node in self.getChildrenNodes():
            other_node.markInvalid(new_value)
            other_node.markDescendantsInvalid(new_value)

    def eval(self, index=0):
        """Evaluate this `Node`. This is supposed to be overriden. See :ref:`evaluation` for more"""
        self.markDirty(False)
        self.markInvalid(False)
        return 0

    def evalChildren(self):
        """Evaluate all children of this `Node`"""
        for node in self.getChildrenNodes():
            node.eval()


    # traversing nodes functions

    def getChildrenNodes(self) -> 'List[Node]':
        """
        Retreive all first-level children connected to this `Node` `Outputs`

        :return: list of `Nodes` connected to this `Node` from all `Outputs`
        :rtype: List[:class:`~nodeeditor.node_node.Node`]
        """
        if self.outputs == []: return []
        other_nodes = []
        for ix in range(len(self.outputs)):
            for edge in self.outputs[ix].edges:
                other_node = edge.getOtherSocket(self.outputs[ix]).node
                other_nodes.append(other_node)
        return other_nodes


    def getInput(self, index:int=0) -> ['Node', None]:
        """
        Get the **first**  `Node` connected to the  Input specified by `index`

        :param index: Order number of the `Input Socket`
        :type index: ``int``
        :return: :class:`~nodeeditor.node_node.Node` which is connected to the specified `Input` or ``None`` if there is no connection of index is out of range
        :rtype: :class:`~nodeeditor.node_node.Node` or ``None``
        """
        try:
            if index<=(len(self.inputs)-1):
                print("indexz=",index,"lenth",len(self.inputs))
                input_socket = self.inputs[index]
                if len(input_socket.edges) == 0: return None
                connecting_edge = input_socket.edges[0]
                other_socket = connecting_edge.getOtherSocket(self.inputs[index])
                return other_socket.node
        except Exception as e:
            dumpException(e)
            return None

    def getInputWithSocket(self, index:int=0) -> [('Node', 'Socket'), (None, None)]:
        """
        Get the **first**  `Node` connected to the Input specified by `index` and the connection `Socket`

        :param index: Order number of the `Input Socket`
        :type index: ``int``
        :return: Tuple containing :class:`~nodeeditor.node_node.Node` and :class:`~nodeeditor.node_socket.Socket` which
            is connected to the specified `Input` or ``None`` if there is no connection of index is out of range
        :rtype: (:class:`~nodeeditor.node_node.Node`, :class:`~nodeeditor.node_socket.Socket`)
        """
        try:
            input_socket = self.inputs[index]
            if len(input_socket.edges) == 0: return None, None
            connecting_edge = input_socket.edges[0]
            other_socket = connecting_edge.getOtherSocket(self.inputs[index])
            return other_socket.node, other_socket
        except Exception as e:
            dumpException(e)
            return None, None

    def getInputWithSocketIndex(self, index:int=0) -> ('Node', int):
        """
        Get the **first**  `Node` connected to the Input specified by `index` and the connection `Socket`

        :param index: Order number of the `Input Socket`
        :type index: ``int``
        :return: Tuple containing :class:`~nodeeditor.node_node.Node` and :class:`~nodeeditor.node_socket.Socket` which
            is connected to the specified `Input` or ``None`` if there is no connection of index is out of range
        :rtype: (:class:`~nodeeditor.node_node.Node`, int)
        """
        try:
            edge = self.inputs[index].edges[0]
            socket = edge.getOtherSocket(self.inputs[index])
            return socket.node, socket.index
        except IndexError:
            # print("EXC: Trying to get input with socket index %d, but none is attached to" % index, self)
            return None, None
        except Exception as e:
            dumpException(e)
            return None, None

    def getInputs(self, index:int=0) -> 'List[Node]':
        """
        Get **all** `Nodes` connected to the Input specified by `index`

        :param index: Order number of the `Input Socket`
        :type index: ``int``
        :return: all :class:`~nodeeditor.node_node.Node` instances which are connected to the specified `Input` or ``[]`` if there is no connection of index is out of range
        :rtype: List[:class:`~nodeeditor.node_node.Node`]
        """
        ins = []
        for edge in self.inputs[index].edges:
            other_socket = edge.getOtherSocket(self.inputs[index])
            ins.append(other_socket.node)
        return ins

    def getOutputs(self, index:int=0) -> 'List[Node]':
        """
        Get **all** `Nodes` connected to the Output specified by `index`

        :param index: Order number of the `Output Socket`
        :type index: ``int``
        :return: all :class:`~nodeeditor.node_node.Node` instances which are connected to the specified `Output` or ``[]`` if there is no connection of index is out of range
        :rtype: List[:class:`~nodeeditor.node_node.Node`]
        """
        outs = []
        for edge in self.outputs[index].edges:
            other_socket = edge.getOtherSocket(self.outputs[index])
            outs.append(other_socket.node)
        return outs


    # serialization functions

    def serialize(self) -> OrderedDict:
        inputs, outputs = [], []
        for socket in self.inputs: inputs.append(socket.serialize())
        for socket in self.outputs: outputs.append(socket.serialize())
        ser_content = self.content.serialize() if isinstance(self.content, Serializable) else {}
        return OrderedDict([
            ('id', self.id),
            ('title', self.title),
            ('pos_x', self.grNode.scenePos().x()),
            ('pos_y', self.grNode.scenePos().y()),
            ('inputs', inputs),
            ('outputs', outputs),
            ('content', ser_content),
        ])

    def deserialize(self, data:dict, hashmap:dict={}, restore_id:bool=True, *args, **kwargs) -> bool:
        try:
            if restore_id: self.id = data['id']
            hashmap[data['id']] = self

            self.setPos(data['pos_x'], data['pos_y'])
            self.title = data['title']

            data['inputs'].sort(key=lambda socket: socket['index'] + socket['position'] * 10000 )
            data['outputs'].sort(key=lambda socket: socket['index'] + socket['position'] * 10000 )
            num_inputs = len( data['inputs'] )
            num_outputs = len( data['outputs'] )

            # print("> deserialize node,   num inputs:", num_inputs, "num outputs:", num_outputs)
            # pp(data)

            # possible way to do it is reuse existing sockets...
            # dont create new ones if not necessary

            for socket_data in data['inputs']:
                found = None
                for socket in self.inputs:
                    # print("\t", socket, socket.index, "=?", socket_data['index'])
                    if socket.index == socket_data['index']:
                        found = socket
                        break
                if found is None:
                    # print("deserialization of socket data has not found input socket with index:", socket_data['index'])
                    # print("actual socket data:", socket_data)
                    # we can create new socket for this
                    found = self.__class__.Socket_class(
                        node=self, index=socket_data['index'], position=socket_data['position'],
                        socket_type=socket_data['socket_type'], count_on_this_node_side=num_inputs,
                        is_input=True
                    )
                    self.inputs.append(found)  # append newly created input to the list
                found.deserialize(socket_data, hashmap, restore_id)


            for socket_data in data['outputs']:
                found = None
                for socket in self.outputs:
                    # print("\t", socket, socket.index, "=?", socket_data['index'])
                    if socket.index == socket_data['index']:
                        found = socket
                        break
                if found is None:
                    # print("deserialization of socket data has not found output socket with index:", socket_data['index'])
                    # print("actual socket data:", socket_data)
                    # we can create new socket for this
                    found = self.__class__.Socket_class(
                        node=self, index=socket_data['index'], position=socket_data['position'],
                        socket_type=socket_data['socket_type'], count_on_this_node_side=num_outputs,
                        is_input=False
                    )
                    self.outputs.append(found)  # append newly created output to the list
                found.deserialize(socket_data, hashmap, restore_id)

        except Exception as e: dumpException(e)

        # also deseralize the content of the node
        # so far the rest was ok, now as last step the content...
        if isinstance(self.content, Serializable):
            res = self.content.deserialize(data['content'], hashmap)
            return res


        return True



import pandas as pd

def below_above_stream(self):
    new_rows = []

    # Division de chaque ligne en deux
    for index, row in self.stream_list.iterrows():
    
        if self.stream_list.loc[index, 'StreamType']=='HS' and self.stream_list.loc[index, 'STo']<self.Pinch_Temperature:
            row_below = row.copy()
            if self.Pinch_Temperature<=row_below['STi']:
                row_below['STi'] = self.Pinch_Temperature
            new_rows.append(row_below)

        if self.stream_list.loc[index, 'StreamType']=='CS' and self.stream_list.loc[index, 'STi']<self.Pinch_Temperature:
            row_below = row.copy()
            if self.Pinch_Temperature<=row_below['STo']:
                row_below['STo'] = self.Pinch_Temperature          
            new_rows.append(row_below)

##################################################################################
        if self.stream_list.loc[index, 'StreamType']=='HS' and self.stream_list.loc[index, 'STi']>self.Pinch_Temperature:
            row_above = row.copy()
            if self.Pinch_Temperature>=self.stream_list.loc[index, 'STo']:
                row_above['STo'] = self.Pinch_Temperature
            else:
                row_above['STo'] = self.stream_list.loc[index, 'STo']
            new_rows.append(row_above)


        if self.stream_list.loc[index, 'StreamType']=='CS' and self.stream_list.loc[index, 'STo']>self.Pinch_Temperature:
            row_above = row.copy()
            if row_above['STi'] >= self.Pinch_Temperature:
                # Tout le flux est au-dessus du pinch -> garder tel quel
                pass
            else:
                # Le flux croise le pinch -> couper STi au pinch
                row_above['STi'] = self.Pinch_Temperature
            new_rows.append(row_above)
        





    

    # Création du nouveau DataFrame
    df_divided = pd.DataFrame(new_rows).sort_values(by=['id', 'STi']).reset_index(drop=True)


    for i, row in df_divided.iterrows():
        if row['StreamType'] == "CS":
            df_divided.at[i, 'Ti'] = row['STi'] - row['dTmin2']
            df_divided.at[i, 'To'] = row['STo'] - row['dTmin2']
        else:  # Pour les flux "HS"
            df_divided.at[i, 'Ti'] = row['STi'] + row['dTmin2']
            df_divided.at[i, 'To'] = row['STo'] + row['dTmin2']
    
    



    # Créer des copies indépendantes pour df_above et df_below
    self.stream_list_above = df_divided[(df_divided['STi'] >= self.Pinch_Temperature) & (df_divided['STo'] >= self.Pinch_Temperature) & (df_divided['STi'] != df_divided['STo'])].copy()
    self.stream_list_below = df_divided[(df_divided['STi'] <= self.Pinch_Temperature) & (df_divided['STo'] <= self.Pinch_Temperature) & (df_divided['STi'] != df_divided['STo'])].copy()

    # Effectuer les modifications avec .loc
    self.stream_list_above.loc[:, 'delta_H'] = self.stream_list_above['mCp'] * (self.stream_list_above['To'] - self.stream_list_above['Ti'])
    self.stream_list_below.loc[:, 'delta_H'] = self.stream_list_below['mCp'] * (self.stream_list_below['To'] - self.stream_list_below['Ti'])



# -*- coding: utf-8 -*-
"""
A module containing Serializable "Interface". We pretend its an abstract class
"""
from collections import OrderedDict

class Serializable():
    def __init__(self):
        """
        Default constructor automatically creates data which are common to any serializable object.
        In our case we create ``self.id`` which we use in every object in NodeEditor.
        """
        self.id = id(self)

    def serialize(self) -> OrderedDict:
        """
        Serialization method to serialize this class data into ``OrderedDict`` which can be stored
        in memory or file easily.

        :return: data serialized in ``OrderedDict``
        :rtype: ``OrderedDict``
        """
        raise NotImplemented()

    def deserialize(self, data:dict, hashmap:dict={}, restore_id:bool=True) -> bool:
        """
        Deserialization method which take data in python ``dict`` format with helping `hashmap` containing
        references to existing entities.

        :param data: Dictionary containing serialized data
        :type data: ``dict``
        :param hashmap: Helper dictionary containing references (by id == key) to existing objects
        :type hashmap: ``dict``
        :param restore_id: True if we are creating new Sockets. False is usefull when loading existing
            Sockets which we want to keep the existing object's `id`.
        :type restore_id: bool
        :return: ``True`` if deserialization was successfull, otherwise ``False``
        :rtype: ``bool``
        """
        raise NotImplemented()



# -*- coding: utf-8 -*-
"""
A module containing all code for working with Clipboard
"""
from collections import OrderedDict
from NodeEditor.nodeeditor.node_graphics_edge import QDMGraphicsEdge
from NodeEditor.nodeeditor.node_edge import Edge


DEBUG = False
DEBUG_PASTING = False


class SceneClipboard():
    """
    Class contains all the code for serialization/deserialization from Clipboard
    """
    def __init__(self, scene:'Scene'):
        """
        :param scene: Reference to the :class:`~nodeeditor.node_scene.Scene`
        :type scene: :class:`~nodeeditor.node_scene.Scene`

        :Instance Attributes:

        - **scene** - reference to the :class:`~nodeeditor.node_scene.Scene`
        """
        self.scene = scene

    def serializeSelected(self, delete:bool=False) -> OrderedDict:
        """
        Serializes selected items in the Scene into ``OrderedDict``

        :param delete: True if you want to delete selected items after serialization. Usefull for Cut operation
        :type delete: ``bool``
        :return: Serialized data of current selection in NodeEditor :class:`~nodeeditor.node_scene.Scene`
        """
        if DEBUG: print("-- COPY TO CLIPBOARD ---")

        sel_nodes, sel_edges, sel_sockets = [], [], {}

        # sort edges and nodes
        for item in self.scene.grScene.selectedItems():
            if hasattr(item, 'node'):
                sel_nodes.append(item.node.serialize())
                for socket in (item.node.inputs + item.node.outputs):
                    sel_sockets[socket.id] = socket
            elif isinstance(item, QDMGraphicsEdge):
                sel_edges.append(item.edge)


        # debug
        if DEBUG:
            print("  NODES\n      ", sel_nodes)
            print("  EDGES\n      ", sel_edges)
            print("  SOCKETS\n     ", sel_sockets)


        # remove all edges which are not connected to a nodeeditor in our list
        edges_to_remove = []
        for edge in sel_edges:
            if edge.start_socket.id in sel_sockets and edge.end_socket.id in sel_sockets:
                # if DEBUG: print(" edge is ok, connected with both sides")
                pass
            else:
                if DEBUG: print("edge", edge, "is not connected with both sides")
                edges_to_remove.append(edge)
        for edge in edges_to_remove:
            sel_edges.remove(edge)

        # make final list of edges
        edges_final = []
        for edge in sel_edges:
            edges_final.append(edge.serialize())

        if DEBUG: print("our final edge list:", edges_final)


        data = OrderedDict([
            ('nodes', sel_nodes),
            ('edges', edges_final),
        ])


        # if CUT (aka delete) remove selected items
        if delete:
            self.scene.getView().deleteSelected()
            # store our history
            self.scene.history.storeHistory("Cut out elements from scene", setModified=True)

        return data

    def deserializeFromClipboard(self, data:dict, *args, **kwargs):
        """
        Deserializes data from Clipboard.

        :param data: ``dict`` data for deserialization to the :class:`nodeeditor.node_scene.Scene`.
        :type data: ``dict``
        """

        hashmap = {}

        # calculate mouse pointer - scene position
        view = self.scene.getView()
        mouse_scene_pos = view.last_scene_mouse_position

        # calculate selected objects bbox and center
        minx, maxx, miny, maxy = 10000000,-10000000, 10000000,-10000000
        for node_data in data['nodes']:
            x, y = node_data['pos_x'], node_data['pos_y']
            if x < minx: minx = x
            if x > maxx: maxx = x
            if y < miny: miny = y
            if y > maxy: maxy = y

        # add width and height of a node
        maxx -= 180
        maxy += 100

        relbboxcenterx = (minx + maxx) / 2 - minx
        relbboxcentery = (miny + maxy) / 2 - miny

        if DEBUG_PASTING:
            print (" *** PASTA:")
            print("Copied boudaries:\n\tX:", minx, maxx, "   Y:", miny, maxy)
            print("\tbbox_center:", relbboxcenterx, relbboxcentery)

        # calculate the offset of the newly creating nodes
        mousex, mousey = mouse_scene_pos.x(), mouse_scene_pos.y()

        # create each node
        created_nodes = []

        self.scene.setSilentSelectionEvents()

        self.scene.doDeselectItems()

        for node_data in data['nodes']:
            new_node = self.scene.getNodeClassFromData(node_data)(self.scene)
            new_node.deserialize(node_data, hashmap, restore_id=False, *args, **kwargs)
            created_nodes.append(new_node)

            # readjust the new nodeeditor's position

            # new node's current position
            posx, posy = new_node.pos.x(), new_node.pos.y()
            newx, newy = mousex + posx - minx, mousey + posy - miny

            new_node.setPos(newx, newy)

            new_node.doSelect()

            if DEBUG_PASTING:
                print("** PASTA SUM:")
                print("\tMouse pos:", mousex, mousey)
                print("\tnew node pos:", posx, posy)
                print("\tFINAL:", newx, newy)

        # create each edge
        if 'edges' in data:
            for edge_data in data['edges']:
                new_edge = Edge(self.scene)
                new_edge.deserialize(edge_data, hashmap, restore_id=False, *args, **kwargs)


        self.scene.setSilentSelectionEvents(False)

        # store history
        self.scene.history.storeHistory("Pasted elements in scene", setModified=True)

        return created_nodes


# Modèle de l'échangeur
from AHU.AirPort.AirPort import AirPort
from AHU.air_humide import air_humide
import pandas as pd

class Object:
    def __init__(self):
        # Initialisation des ports d'entrée et de sortie
        self.id=2 #identifiant de l'objet
        self.Inlet1 = AirPort()  # Air frais entrant
        self.Outlet1 = AirPort()  # Air frais sortant
        self.Inlet2 = AirPort()  # Air extrait entrant
        self.Outlet2 = AirPort()  # Air extrait sortant

        # Paramètres de l'échangeur
        self.h_efficiency = None  # Efficacité enthalpique (calculée)
        self.T_efficiency = 80    # Efficacité en température (%)
        self.T_efficiency_target=None
        self.T_target=16

        self.heat_transfer = None  # Transfert de chaleur (W)
        self.T1o = None  # Température de sortie de l'air frais (°C)
        self.df=None
        self.F_dry_min=None

    def calculate(self):
        # Détermination du débit massique minimum
        self.F_dry_min = min(self.Inlet1.F_dry, self.Inlet2.F_dry)
        
        # Cas hivernal : l'air extrait est plus chaud que l'air extérieur
        if (self.Inlet2.T >= self.Inlet1.T and self.Inlet1.T<=self.T_target):
            print("hiver")

            self.T_efficiency_target=100*self.Inlet1.F_dry*(self.T_target-self.Inlet1.T)/(self.F_dry_min*(self.Inlet2.T - self.Inlet1.T))
            
            if self.T_efficiency>self.T_efficiency_target:
                self.T_efficiency=max(self.T_efficiency_target,0)
            


            # Calcul de la température de sortie de l'air frais
            self.T1o = self.Inlet1.T + (self.F_dry_min*(self.T_efficiency / 100) * (self.Inlet2.T - self.Inlet1.T))/self.Inlet1.F_dry
          

            # Calcul de l'enthalpie de sortie de l'air frais
            self.Outlet1.h = air_humide.Air_h(T_db=self.T1o, w=self.Inlet1.w)
            self.Outlet1.w = self.Inlet1.w  # L'humidité absolue reste constante

            # Calcul du transfert de chaleur
            self.heat_transfer = (self.Outlet1.h - self.Inlet1.h) * self.Inlet1.F_dry

            # Calcul de l'efficacité enthalpique
            self.h_efficiency = 100 * self.Inlet1.F_dry * (self.Outlet1.h - self.Inlet1.h) / (self.F_dry_min * (self.Inlet2.h - self.Inlet1.h))

            # Calcul des propriétés de sortie de l'air extrait
            self.Outlet2.h = -(self.heat_transfer / self.Inlet2.F_dry) + self.Inlet2.h
            
            # Calcul de la température de rosée avant condensation
            self.Tdp_dry = air_humide.Air_T_dp(w=self.Inlet2.w)
       
            
            # Calcul de la température de rosée selon l'enthalpie de sortie d'air vicié
            self.Tdp_wet = air_humide.Air_T_dp(h=self.Outlet2.h)
 
            
            # Vérification de la condensation
            if self.Tdp_wet <= self.Tdp_dry:
                print("condensation")
                self.Outlet2.w = air_humide.Air_w(h=self.Outlet2.h, RH=100)
            else:
                print("pas de condensation")
                self.Outlet2.w = self.Inlet2.w

        elif (self.Inlet2.T >= self.Inlet1.T and self.Inlet1.T>self.T_target):
            self.Outlet1.h = self.Inlet1.h # ne rien échanger
            self.Outlet1.w = self.Inlet1.w  
           
        # Cas estival : l'air extérieur est plus chaud que l'air extrait
        elif (self.Inlet2.T < self.Inlet1.T):
            print("été")
            self.T1o = self.Inlet1.T - (self.F_dry_min*(self.T_efficiency / 100) * (self.Inlet1.T-self.Inlet2.T))/self.Inlet1.F_dry
            # Calcul de l'enthalpie de sortie de l'air frais
            self.Outlet1.h = air_humide.Air_h(T_db=self.T1o, w=self.Inlet1.w)
            self.Outlet1.w = self.Inlet1.w  # L'humidité absolue reste constante
            print(f"self.T1i={self.Inlet1.T}")
            print(f"self.T1o={self.T1o}")

            # Calcul du transfert de chaleur
            self.heat_transfer = (self.Outlet1.h - self.Inlet1.h) * self.Inlet1.F_dry

            # Calcul de l'efficacité enthalpique
            self.h_efficiency = 100 * self.Inlet1.F_dry * (self.Outlet1.h - self.Inlet1.h) / (self.F_dry_min * (self.Inlet2.h - self.Inlet1.h))

            # Calcul des propriétés de sortie de l'air extrait
            self.Outlet2.h = -(self.heat_transfer / self.Inlet2.F_dry) + self.Inlet2.h
            self.Outlet2.w=self.Inlet2.w

        else:
            # Calcul de l'enthalpie de sortie de l'air frais
            pass
   
        # Conservation du débit massique et de l'humidité absolue
        self.Outlet1.F_dry = self.Inlet1.F_dry
        self.Outlet1.w = self.Inlet1.w
        self.Outlet2.F_dry = self.Inlet2.F_dry

        self.Outlet1.update_properties()
        self.Outlet2.update_properties()

        data = {
            'ID': [self.id],  # Identifiant de l'objet

            'Heat recovery (kW)': [self.heat_transfer],
            'sefficacité échangeur T_efficiency (%)': [self.T_efficiency],
            'rendement thermique (rapport des enthalpies) h_efficiency (%)': [self.h_efficiency],

            'self.Inlet1.T (C)': [self.Inlet1.T],  # Température de l'air
            'self.Outlet1.T (C)': [self.Outlet1.T],  # Température de l'air
            'self.Inlet2.T (C)': [self.Inlet2.T],  # Température de l'air
            'self.Outlet2.T (C)': [self.Outlet2.T],  # Température de l'air

            'self.Inlet1.RH (%)': [self.Inlet1.RH],  # Humidité relative
            'self.Outlet1.RH (%)': [self.Outlet1.RH],  # Humidité relative
            'self.Inlet2.RH (%)': [self.Inlet2.RH],  # Humidité relative
            'self.Outlet2.RH (%)': [self.Outlet2.RH],  # Humidité relative


            'self.Inlet1.F (kg/s)': [self.Inlet1.F],  # Débit d'air en kg/s
            'self.Inlet1.F_dry (kg/s)': [self.Inlet1.F_dry],  # Débit d'air sec en kg/s
            'self.Outlet1.F (kg/s)': [self.Outlet1.F],  # Débit d'air en kg/s
            'self.Outlet1.F_dry (kg/s)': [self.Outlet1.F_dry],  # Débit d'air sec en kg/s
            'self.Inlet2.F (kg/s)': [self.Inlet2.F],  # Débit d'air en kg/s
            'self.Inlet2.F_dry (kg/s)': [self.Inlet2.F_dry],  # Débit d'air sec en kg/s
            'self.Outlet2.F (kg/s)': [self.Outlet2.F],  # Débit d'air en kg/s
            'self.Outlet2.F_dry (kg/s)': [self.Outlet2.F_dry],  # Débit d'air sec en kg/s

            'self.Inlet1.P (Pa)': [self.Inlet1.P],  # Pression de l'air
            'self.Outlet1.P (Pa)': [self.Outlet1.P],  # Pression de l'air
            'self.Inlet2.P (Pa)': [self.Inlet2.P],  # Pression de l'air
            'self.Outlet2.P (Pa)': [self.Outlet2.P],  # Pression de l'air

            'self.Inlet1.h (kJ/kg)': [self.Inlet1.h],  # Enthalpie spécifique
            'self.Outlet1.h (kJ/kg)': [self.Outlet1.h],  # Enthalpie spécifique
            'self.Inlet2.h (kJ/kg)': [self.Inlet2.h],  # Enthalpie spécifique
            'self.Outlet2.h (kJ/kg)': [self.Outlet2.h],  # Enthalpie spécifique

            'self.Inlet1.w (gH2O/kgdry)': [self.Inlet1.w],  # Humidité absolue
            'self.Outlet1.w (gH2O/kgdry)': [self.Outlet1.w],  # Humidité absolue
            'self.Inlet2.w (gH2O/kgdry)': [self.Inlet2.w],  # Humidité absolue
           'self.Outlet2.w (gH2O/kgdry)': [self.Outlet2.w],  # Humidité absolue


            'self.Inlet1.Pv_sat (Pa)': [self.Inlet1.Pv_sat],  # Pression de vapeur saturante
            'self.Outlet1.Pv_sat (Pa)': [self.Outlet1.Pv_sat],  # Pression de vapeur saturante
            'self.Inlet2.Pv_sat (Pa)': [self.Inlet2.Pv_sat],  # Pression de vapeur saturante
            'self.Outlet2.Pv_sat (Pa)': [self.Outlet2.Pv_sat],  # Pression de vapeur saturante
        }

        # Convertir les données en DataFrame et transposer pour avoir les paramètres en lignes
        self.df = pd.DataFrame(data).T
       


# -*- coding: utf-8 -*-
"""
A module containing Graphic representation of :class:`~nodeeditor.node_scene.Scene`
"""
import math
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *


class QDMGraphicsScene(QGraphicsScene):
    """Class representing Graphic of :class:`~nodeeditor.node_scene.Scene`"""
    #: pyqtSignal emitted when some item is selected in the `Scene`
    itemSelected = pyqtSignal()
    #: pyqtSignal emitted when items are deselected in the `Scene`
    itemsDeselected = pyqtSignal()

    def __init__(self, scene:'Scene', parent:QWidget=None):
        """
        :param scene: reference to the :class:`~nodeeditor.node_scene.Scene`
        :type scene: :class:`~nodeeditor.node_scene.Scene`
        :param parent: parent widget
        :type parent: QWidget
        """
        super().__init__(parent)

        self.scene = scene

        # settings
        self.gridSize = 20
        self.gridSquares = 5

        self.initAssets()
        self.setBackgroundBrush(self._color_background)

    def initAssets(self):
        """Initialize ``QObjects`` like ``QColor``, ``QPen`` and ``QBrush``"""
        self._color_background = QColor("#393939")
        self._color_light = QColor("#2f2f2f")
        self._color_dark = QColor("#292929")

        self._pen_light = QPen(self._color_light)
        self._pen_light.setWidth(1)
        self._pen_dark = QPen(self._color_dark)
        self._pen_dark.setWidth(2)


    # the drag events won't be allowed until dragMoveEvent is overriden
    def dragMoveEvent(self, event):
        """Overriden Qt's dragMoveEvent to enable Qt's Drag Events"""
        pass

    def setGrScene(self, width:int, height:int):
        """Set `width` and `height` of the `Graphics Scene`"""
        self.setSceneRect(-width // 2, -height // 2, width, height)

    def drawBackground(self, painter:QPainter, rect:QRect):
        """Draw background scene grid"""
        super().drawBackground(painter, rect)

        # here we create our grid
        left = int(math.floor(rect.left()))
        right = int(math.ceil(rect.right()))
        top = int(math.floor(rect.top()))
        bottom = int(math.ceil(rect.bottom()))

        first_left = left - (left % self.gridSize)
        first_top = top - (top % self.gridSize)

        # compute all lines to be drawn
        lines_light, lines_dark = [], []
        for x in range(first_left, right, self.gridSize):
            if (x % (self.gridSize*self.gridSquares) != 0): lines_light.append(QLine(x, top, x, bottom))
            else: lines_dark.append(QLine(x, top, x, bottom))

        for y in range(first_top, bottom, self.gridSize):
            if (y % (self.gridSize*self.gridSquares) != 0): lines_light.append(QLine(left, y, right, y))
            else: lines_dark.append(QLine(left, y, right, y))


        # draw the lines
        painter.setPen(self._pen_light)
        painter.drawLines(*lines_light)

        painter.setPen(self._pen_dark)
        painter.drawLines(*lines_dark)


# -*- coding: utf-8 -*-
"""
A module containing Graphics representation of Edge
"""
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *

from NodeEditor.nodeeditor.node_graphics_edge_path import GraphicsEdgePathBezier, GraphicsEdgePathDirect


class QDMGraphicsEdge(QGraphicsPathItem):
    """Base class for Graphics Edge"""
    def __init__(self, edge:'Edge', parent:QWidget=None):
        """
        :param edge: reference to :class:`~nodeeditor.node_edge.Edge`
        :type edge: :class:`~nodeeditor.node_edge.Edge`
        :param parent: parent widget
        :type parent: ``QWidget``

        :Instance attributes:

            - **edge** - reference to :class:`~nodeeditor.node_edge.Edge`
            - **posSource** - ``[x, y]`` source position in the `Scene`
            - **posDestination** - ``[x, y]`` destination position in the `Scene`
        """
        super().__init__(parent)

        self.edge = edge

        # create instance of our path class
        self.pathCalculator = self.determineEdgePathClass()(self)

        # init our flags
        self._last_selected_state = False
        self.hovered = False

        # init our variables
        self.posSource = [0, 0]
        self.posDestination = [200, 100]

        self.initAssets()
        self.initUI()

    def initUI(self):
        """Set up this ``QGraphicsPathItem``"""
        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.setAcceptHoverEvents(True)
        self.setZValue(-1)

    def initAssets(self):
        """Initialize ``QObjects`` like ``QColor``, ``QPen`` and ``QBrush``"""
        self._color = self._default_color = QColor("#001000")
        self._color_selected = QColor("#00ff00")
        self._color_hovered = QColor("#FF37A6FF")
        self._pen = QPen(self._color)
        self._pen_selected = QPen(self._color_selected)
        self._pen_dragging = QPen(self._color)
        self._pen_hovered = QPen(self._color_hovered)
        self._pen_dragging.setStyle(Qt.DashLine)
        self._pen.setWidthF(3.0)
        self._pen_selected.setWidthF(3.0)
        self._pen_dragging.setWidthF(3.0)
        self._pen_hovered.setWidthF(5.0)

    def createEdgePathCalculator(self):
        """Create instance of :class:`~nodeeditor.node_graphics_edge_path.GraphicsEdgePathBase`"""
        self.pathCalculator = self.determineEdgePathClass()(self)
        return self.pathCalculator

    def determineEdgePathClass(self):
        """Decide which GraphicsEdgePath class should be used to calculate path according to edge.edge_type value"""
        from NodeEditor.nodeeditor.node_edge import EDGE_TYPE_BEZIER, EDGE_TYPE_DIRECT
        if self.edge.edge_type == EDGE_TYPE_BEZIER:
            return GraphicsEdgePathBezier
        if self.edge.edge_type == EDGE_TYPE_DIRECT:
            return GraphicsEdgePathDirect
        else:
            return GraphicsEdgePathBezier

    def makeUnselectable(self):
        """Used for drag edge to disable click detection over this graphics item"""
        self.setFlag(QGraphicsItem.ItemIsSelectable, False)
        self.setAcceptHoverEvents(False)

    def changeColor(self, color):
        """Change color of the edge from string hex value '#00ff00'"""
        # print("^Called change color to:", color.red(), color.green(), color.blue(), "on edge:", self.edge)
        self._color = QColor(color) if type(color) == str else color
        self._pen = QPen(self._color)
        self._pen.setWidthF(3.0)

    def setColorFromSockets(self) -> bool:
        """Change color according to connected sockets. Returns ``True`` if color can be determined"""
        socket_type_start = self.edge.start_socket.socket_type
        socket_type_end = self.edge.end_socket.socket_type
        if socket_type_start != socket_type_end: return False
        self.changeColor(self.edge.start_socket.grSocket.getSocketColor(socket_type_start))

    def onSelected(self):
        """Our event handling when the edge was selected"""
        self.edge.scene.grScene.itemSelected.emit()

    def doSelect(self, new_state:bool=True):
        """Safe version of selecting the `Graphics Node`. Takes care about the selection state flag used internally

        :param new_state: ``True`` to select, ``False`` to deselect
        :type new_state: ``bool``
        """
        self.setSelected(new_state)
        self._last_selected_state = new_state
        if new_state: self.onSelected()

    def mouseReleaseEvent(self, event):
        """Overriden Qt's method to handle selecting and deselecting this `Graphics Edge`"""
        super().mouseReleaseEvent(event)
        if self._last_selected_state != self.isSelected():
            self.edge.scene.resetLastSelectedStates()
            self._last_selected_state = self.isSelected()
            self.onSelected()

    def hoverEnterEvent(self, event: 'QGraphicsSceneHoverEvent') -> None:
        """Handle hover effect"""
        self.hovered = True
        self.update()

    def hoverLeaveEvent(self, event: 'QGraphicsSceneHoverEvent') -> None:
        """Handle hover effect"""
        self.hovered = False
        self.update()

    def setSource(self, x:float, y:float):
        """ Set source point

        :param x: x position
        :type x: ``float``
        :param y: y position
        :type y: ``float``
        """
        self.posSource = [x, y]

    def setDestination(self, x:float, y:float):
        """ Set destination point

        :param x: x position
        :type x: ``float``
        :param y: y position
        :type y: ``float``
        """
        self.posDestination = [x, y]

    def boundingRect(self) -> QRectF:
        """Defining Qt' bounding rectangle"""
        return self.shape().boundingRect()

    def shape(self) -> QPainterPath:
        """Returns ``QPainterPath`` representation of this `Edge`

        :return: path representation
        :rtype: ``QPainterPath``
        """
        return self.calcPath()

    def paint(self, painter, QStyleOptionGraphicsItem, widget=None):
        """Qt's overriden method to paint this Graphics Edge. Path calculated in :func:`~nodeeditor.node_graphics_edge.QDMGraphicsEdge.calcPath` method"""
        self.setPath(self.calcPath())

        painter.setBrush(Qt.NoBrush)

        if self.hovered and self.edge.end_socket is not None:
            painter.setPen(self._pen_hovered)
            painter.drawPath(self.path())

        if self.edge.end_socket is None:
            painter.setPen(self._pen_dragging)
        else:
            painter.setPen(self._pen if not self.isSelected() else self._pen_selected)

        painter.drawPath(self.path())

    def intersectsWith(self, p1:QPointF, p2:QPointF) -> bool:
        """Does this Graphics Edge intersect with line between point A and point B ?

        :param p1: point A
        :type p1: ``QPointF``
        :param p2: point B
        :type p2: ``QPointF``
        :return: ``True`` if this `Graphics Edge` intersects
        :rtype: ``bool``
        """
        cutpath = QPainterPath(p1)
        cutpath.lineTo(p2)
        path = self.calcPath()
        return cutpath.intersects(path)

    def calcPath(self) -> QPainterPath:
        """Will handle drawing QPainterPath from Point A to B. Internally there exist self.pathCalculator which
        is an instance of derived :class:`~nodeeditor.node_graphics_edge_path.GraphicsEdgePathBase` class
        containing the actual `calcPath()` function - computing how the edge should look like.

        :returns: ``QPainterPath`` of the edge connecting `source` and `destination`
        :rtype: ``QPainterPath``
        """
        return self.pathCalculator.calcPath()




import pvlib
import pandas as pd
import matplotlib.pyplot as plt

class SolarSystem:
    def __init__(self, latitude, longitude, name, altitude, timezone, azimut, inclinaison):
        self.latitude = latitude
        self.longitude = longitude
        self.name = name
        self.altitude = altitude
        self.timezone = timezone
        self.azimut = azimut
        self.inclinaison = inclinaison
        self.weather = None
        self.module = None
        self.inverter = None
        self.temperature_model_parameters = None
        self.system = None
        self.solpos = None
        self.dni_extra = None
        self.airmass = None
        self.pressure = None
        self.am_abs = None
        self.aoi = None
        self.total_irradiance = None
        self.cell_temperature = None
        self.effective_irradiance = None
        self.dc = None
        self.ac = None
        self.annual_energy = None

    def retrieve_module_inverter_data(self):
        # Récupération de la base de données des modules Sandia
        sandia_modules = pvlib.pvsystem.retrieve_sam('SandiaMod')
        #print("Modules PV Sandia", sandia_modules)
        # Récupération de la base de données des onduleurs SAPM
        sapm_inverters = pvlib.pvsystem.retrieve_sam('cecinverter')
        #print("Onduleurs", sapm_inverters)
        # Sélection d'un module spécifique à partir de la base de données des modules Sandia
        self.module = sandia_modules['Canadian_Solar_CS5P_220M___2009_']
        # Sélection d'un onduleur spécifique à partir de la base de données des onduleurs SAPM
        self.inverter = sapm_inverters['ABB__MICRO_0_25_I_OUTD_US_208__208V_']
        # Récupération des paramètres du modèle de température pour le modèle SAPM et le module open rack glass-glass
        self.temperature_model_parameters = pvlib.temperature.TEMPERATURE_MODEL_PARAMETERS['sapm']['open_rack_glass_glass']
        #print("Modèle de température", self.temperature_model_parameters)

    def retrieve_weather_data(self):
        # Récupération des données météorologiques TMY à partir de PVGIS pour l'emplacement spécifié
        weather = pvlib.iotools.get_pvgis_tmy(self.latitude, self.longitude, map_variables=True)[0]
        weather.index.name = "utc_time"  # Définition du nom de l'index des données météorologiques
        self.weather = weather
        #print("Données météorologiques typiques :", self.weather)

    def calculate_solar_parameters(self):
        # Configuration du système solaire avec le module, l'onduleur et l'azimut de surface fixé à 180 degrés
        print("azimut:",self.azimut)
        self.system = {'module': self.module, 'inverter': self.inverter, 'surface_azimuth': self.azimut}
        # Définition de l'inclinaison de la surface du système solaire à la latitude de l'emplacement
        self.system['surface_tilt'] = self.inclinaison
        print("inclinaison", self.system['surface_tilt'])
        # Calcul de la position solaire
        self.solpos = pvlib.solarposition.get_solarposition(
            time=self.weather.index,
            latitude=self.latitude,
            longitude=self.longitude,
            altitude=self.altitude,
            temperature=self.weather["temp_air"],
            pressure=pvlib.atmosphere.alt2pres(self.altitude),
        )
        #print("position du soleil===", self.solpos)
        # Calcul du rayonnement solaire supplémentaire
        self.dni_extra = pvlib.irradiance.get_extra_radiation(self.weather.index)
        #print("rayonnement solaire suplémentaire", self.dni_extra)
        # Calcul de la masse d'air relative
        self.airmass = pvlib.atmosphere.get_relative_airmass(self.solpos['apparent_zenith'])
        #print("masse d'air relative", self.airmass)
        # Calcul de la pression atmosphérique
        self.pressure = pvlib.atmosphere.alt2pres(self.altitude)
        #print("pression atm", self.pressure)
        # Calcul de la masse d'air absolue
        self.am_abs = pvlib.atmosphere.get_absolute_airmass(self.airmass, self.pressure)
        #print("masse d'air absolue", self.am_abs)
        # Calcul de l'angle d'incidence
        self.aoi = pvlib.irradiance.aoi(
            self.system['surface_tilt'],
            self.system['surface_azimuth'],
            self.solpos["apparent_zenith"],
            self.solpos["azimuth"],
        )
        #print("angle d'incidence", self.aoi)
        # Calcul du rayonnement solaire total
        self.total_irradiance = pvlib.irradiance.get_total_irradiance(
            self.system['surface_tilt'],
            self.system['surface_azimuth'],
            self.solpos['apparent_zenith'],
            self.solpos['azimuth'],
            self.weather['dni'],
            self.weather['ghi'],
            self.weather['dhi'],
            dni_extra=self.dni_extra,
            model='haydavies',
        )
        #print("rayonnement solaire total", self.total_irradiance)
        # Calcul de la température des cellules solaires
        self.cell_temperature = pvlib.temperature.sapm_cell(
            self.total_irradiance['poa_global'],
            self.weather["temp_air"],
            self.weather["wind_speed"],
            **self.temperature_model_parameters,
        )
        #print("température des cellules solaires", self.cell_temperature)
        # Calcul du rayonnement solaire effectif
        self.effective_irradiance = pvlib.pvsystem.sapm_effective_irradiance(
            self.total_irradiance['poa_direct'],
            self.total_irradiance['poa_diffuse'],
            self.am_abs,
            self.aoi,
            self.module,
        )
        #print("rayonnement solaire effectif", self.effective_irradiance)
        # Calcul de la puissance CC (courant continu)
        self.dc = pvlib.pvsystem.sapm(self.effective_irradiance, self.cell_temperature, self.module)
        #print("puissance CC", self.dc)
        # Calcul de la puissance CA (courant alternatif)
        self.ac = pvlib.inverter.sandia(self.dc['v_mp'], self.dc['p_mp'], self.inverter)
        #print("puissance CA (courant alternatif)", self.ac)
        # Calcul de l'énergie annuelle
        self.annual_energy = -self.ac.sum()
        print("énergie annuelle:")
        print(self.annual_energy)
        

    def plot_annual_energy(self):
        # Stockage de l'énergie annuelle dans un dictionnaire avec le nom de l'emplacement comme clé
        energies = {self.name: self.annual_energy}
        print("energies résult", energies)
        # Conversion du dictionnaire en une série pandas
        energies = pd.Series(energies)
        # Affichage de la série d'énergies
        print("Rendement énergétique annuel (KWh)", energies / 1000)
        # Affichage du graphique en barres des rendements énergétiques annuels
        energies.plot(kind='bar', rot=0)
        # Configuration de l'axe des ordonnées
        plt.ylabel('Rendement énergétique annuel (Wh)')


import matplotlib.pyplot as plt
import json


import sys
sys.path.append("..") # Adds higher directory to python Modules_CTA path.

import sqlite3

from OpenWeatherMap import get_weather
from get_weather import *

import datetime as dt


def OpenWeatherMap():
    T=0
# Save the current time to a variable ('t')
    
    t = dt.datetime.now()
    


    while True:
        delta = dt.datetime.now()-t
        
        if delta.seconds >= 10:
            #print("+5 secondes")
            # Update 't' variable to new time
            t = dt.datetime.now()
            #print(t)

##################################################################################
            conn = sqlite3.connect('BD_Meteo.db')
            c = conn.cursor()

# Create table
            c.execute('''CREATE TABLE if not exists Eaubonne
             (date text, Temperature real, Humidity real)''')
#####################################données météo##############################################################
# récupérer les données météo
            
            api_key = get_api_key()
#    print(api_key)
    
            location=get_location()
#    print(location[0])

            weather = get_weather(api_key, location[0],location[1],location[2])
# 
            #print(weather['main']['temp']-273.15,"°C",weather['main']['humidity'],"%")
            RH=weather['main']['humidity']
            T=weather['main']['temp']-273.15
            # print(weather)
            print("T = ",T," RH = ",RH)
            with open('data.json', 'w') as outfile:
                json.dump(weather, outfile)
################################################################################################    
# Insert a row of data
            dataForme_insert="INSERT INTO Eaubonne VALUES (?,?,?)"
            #import random
        
            #data_insert=(t,random.randint(-5, 35))
            data_insert=(t,weather['main']['temp']-273.15,weather['main']['humidity'])
            
            
            c.execute(dataForme_insert,data_insert)
            
            #plt.plot(t,weather['main']['temp']-273.15)
            #plt.show()
    

# Save (commit) the changes
            conn.commit()

# We can also close the connection if we are done with it.
# Just be sure any changes have been committed or they will be lost.
            conn.close()
            
<<<<<<< HEAD
       # return T
=======
    #    return T
>>>>>>> c751d069316d9b805d3b17b04797fbd1bc04b23d
            
            
  

OpenWeatherMap()


# -*- coding: utf-8 -*-
"""
A module containing the Edge Validator functions which can be registered as callbacks to
:class:`~nodeeditor.node_edge.Edge` class.

Example of registering Edge Validator callbacks:
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

You can register validation callbacks once for example on the bottom of node_edge.py file or on the
application start with calling this:

.. code-block:: python

    from nodeeditor.node_edge_validators import *

    Edge.registerEdgeValidator(edge_validator_debug)
    Edge.registerEdgeValidator(edge_cannot_connect_two_outputs_or_two_inputs)
    Edge.registerEdgeValidator(edge_cannot_connect_input_and_output_of_same_node)

"""

DEBUG = False


def print_error(*args):
    """Helper method which prints to console if `DEBUG` is set to `True`"""
    if DEBUG: print("Edge Validation Error:", *args)

def edge_validator_debug(input: 'Socket', output: 'Socket') -> bool:
    """This will consider edge always valid, however writes bunch of debug stuff into console"""
    print("VALIDATING:")
    print(input, "input" if input.is_input else "output",  "of node", input.node)
    for s in input.node.inputs+input.node.outputs: print("\t", s, "input" if s.is_input else "output")
    print(output, "input" if input.is_input else "output", "of node", output.node)
    for s in output.node.inputs+output.node.outputs: print("\t", s, "input" if s.is_input else "output")

    return True

def edge_cannot_connect_two_outputs_or_two_inputs(input: 'Socket', output: 'Socket') -> bool:
    """Edge is invalid if it connects 2 output sockets or 2 input sockets"""
    if input.is_output and output.is_output:
        print_error("Connecting 2 outputs")
        return False

    if input.is_input and output.is_input:
        print_error("Connecting 2 inputs")
        return False

    return True

def edge_cannot_connect_input_and_output_of_same_node(input: 'Socket', output:'Socket') -> bool:
    """Edge is invalid if it connects the same node"""
    if input.node == output.node:
        print_error("Connecting the same node")
        return False

    return True


from ThermodynamicCycles.FluidPort.FluidPort import FluidPort
from CoolProp.CoolProp import PropsSI
import pandas as pd 
from datetime import datetime

class Object:
    def __init__(self):
        self.Timestamp=datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.Inlet=FluidPort() 
        self.Outlet=FluidPort()

        self.To=None
        self.P_drop=0
        self.Qth=None
     
        self.df = pd.DataFrame()
        
    def calculate (self):
    
        self.Outlet.P=self.Inlet.P-self.P_drop
        self.Outlet.F=self.Inlet.F
        self.Outlet.fluid=self.Inlet.fluid

        if self.Outlet.F is not None and self.To is not None and self.Qth is None:
            print("Calcul de la puissance thermique ou frigorifique:")
            self.Outlet.T=self.To+273.15
            self.Outlet.h = PropsSI('H','P',self.Outlet.P,'T',self.Outlet.T,self.Outlet.fluid)
            #calcul de la puissance thermique de l'échangeur
            self.Qth=self.Inlet.F*(self.Outlet.h-self.Inlet.h)
        
        if self.Qth is not None and self.To is not None:
            print("calcul du débit!")
            self.Outlet.T=self.To+273.15
            self.Outlet.h = PropsSI('H','P',self.Outlet.P,'T',self.Outlet.T,self.Outlet.fluid)
            print('self.Qth;',self.Qth)
            if (self.Outlet.h-self.Inlet.h)!=0:
                self.Inlet.F=self.Qth/(self.Outlet.h-self.Inlet.h)
            else:
               self.Inlet.F=0.0
            self.Outlet.F=self.Inlet.F
            print("self.Inlet.F:",self.Inlet.F)

        if self.Outlet.F is not None and self.Qth is not None and self.To is None:
            print("calcul de la température de sortie")
            self.Outlet.h=(self.Qth+self.Inlet.F*self.Inlet.h)/self.Inlet.F
            self.To = PropsSI('T','P',self.Outlet.P,'H',self.Outlet.h,self.Outlet.fluid)-273.15



        self.Outlet.S = PropsSI('S','P',self.Outlet.P,'H',self.Outlet.h,self.Inlet.fluid)
        self.Outlet.calculate_properties()

        self.df = pd.DataFrame({'Simple_HEX': [self.Timestamp,self.To,self.Qth/1000,], },
                      index = ['Timestamp','Simple_HEX(°C)','hex_Qhex(kW)'])     


from ThermodynamicCycles.FluidPort.FluidPort import FluidPort
from CoolProp.CoolProp import PropsSI
import pandas as pd 
from datetime import datetime

import math

class Object:
    def __init__(self):
        self.Timestamp=datetime.now().strftime("%Y-%m-%d %H:%M:%S")
   
        self.Inlet1 = FluidPort()
        self.Outlet1 = FluidPort()
        self.Inlet2 = FluidPort()
        self.Outlet2 = FluidPort()

        self.T1o=None
        self.T2o=None
        self.C1=None
        self.C2=None
        self.R=None # Le rapport des débits de capacités des deux fluides est 
        self.Eff=None # efficacité
        self.UA=None
        self.NUT=None
        self.Cmin=None

        self.P_drop=0
   
     
        self.df = pd.DataFrame()


        
    def calculate (self):
    
        self.Outlet1.P=self.Inlet1.P-self.P_drop
        self.Outlet1.F=self.Inlet1.F
        self.Outlet2.F=self.Inlet2.F
        self.Outlet1.fluid=self.Inlet1.fluid
        self.Outlet2.fluid=self.Inlet2.fluid

        #self.Outlet1.calculate_properties()

   


    

        self.C1=self.Inlet1.F*self.Inlet1.cp
        self.C2=self.Inlet2.F*self.Inlet2.cp
        self.Cmin=min(self.C1,self.C2)
        self.Cmax=max(self.C1,self.C2)

        self.R=self.Cmin/self.Cmax


        self.NUT=self.UA/self.Cmin
        #print('self.NUT================',self.NUT)
        self.Eff=(1-math.exp(-self.NUT*(1-self.R)))/(1-self.R*math.exp(-self.NUT*(1-self.R)))
        #print('self.Eff================',self.Eff)
        self.Qth=self.Eff*self.Cmin*(self.Inlet1.T-self.Inlet2.T)
        #print('self.Qth================',self.Qth)

        self.Inlet1.calculate_properties()
        self.Outlet1.h=-(self.Qth/self.Inlet1.F)+self.Inlet1.h
        self.Inlet2.calculate_properties()
        self.Outlet2.h=(self.Qth/self.Inlet2.F)+self.Inlet2.h
        
        

        self.Outlet1.calculate_properties()
        #print('self.Outlet1.T °C================',self.Outlet1.T-273.15)
        self.Outlet2.calculate_properties()
        #print('self.Outlet2.T °C================',self.Outlet2.T-273.15)

        self.df = pd.DataFrame({'DTLM_HEX': [self.Timestamp,self.Qth], },
                      index = ['Timestamp',self.Qth])     


# from ThermodynamicCycles.Source import Source
# from ThermodynamicCycles.Connect import Fluid_connect

# SOURCE1=Source.Object()
# SOURCE1.F_m3h=15
# SOURCE1.Pi_bar=1
# SOURCE1.fluid='Dodecane'
# SOURCE1.Ti_degC=120
# SOURCE1.calculate()

# SOURCE2=Source.Object()
# SOURCE2.F_m3h=14*3.6
# SOURCE2.Pi_bar=1
# SOURCE2.fluid='water'
# SOURCE2.Ti_degC=20
# SOURCE2.calculate()

# HEX=Object()
# HEX.UA=6698.371240548343

# Fluid_connect(HEX.Inlet1,SOURCE1.Outlet)
# Fluid_connect(HEX.Inlet2,SOURCE2.Outlet)
# HEX.calculate()
# print("-------------------------------",HEX.df)

# print(HEX.Outlet1.df)
# print(HEX.Outlet2.df)

# print(HEX.R)
# print(HEX.Eff)



"""
Peng-Robinson EOS - From-scratch implementation with kij.

Uses thermo library only for critical properties (Tc, Pc, omega) and
ideal gas Cp correlations. All EOS calculations (mixing rules, cubic
solver, fugacity coefficients, departure enthalpies) are from scratch
with binary interaction parameters from kij_peng_robinson.

Reference: Chapitre 5 - Equation d'etat de Peng-Robinson
"""

import os
import sys
import numpy as np

_src_dir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
if _src_dir not in sys.path:
    sys.path.insert(0, _src_dir)

from thermo import ChemicalConstantsPackage
from kij_peng_robinson import build_kij_matrix, COMP_INDEX

R = 8.314  # J/(mol·K)
SQRT2 = np.sqrt(2.0)

_NAME_MAP = {
    "C1": "methane", "C2": "ethane", "C3": "propane",
    "iC4": "isobutane", "nC4": "butane",
    "iC5": "isopentane", "nC5": "pentane",
    "C6": "hexane", "C7": "heptane", "C8": "octane",
    "C9": "nonane", "C10": "decane", "C11": "undecane", "C12": "dodecane",
    "CO2": "carbon dioxide", "N2": "nitrogen",
}

COMPONENT_DB = {
    "C1": {"Mw": 16.04}, "C2": {"Mw": 30.07}, "C3": {"Mw": 44.10},
    "iC4": {"Mw": 58.12}, "nC4": {"Mw": 58.12},
    "iC5": {"Mw": 72.15}, "nC5": {"Mw": 72.15},
    "C6": {"Mw": 86.18}, "C7": {"Mw": 100.21}, "C8": {"Mw": 114.23},
    "C9": {"Mw": 128.25}, "C10": {"Mw": 142.28}, "C11": {"Mw": 156.31},
    "C12": {"Mw": 170.34}, "CO2": {"Mw": 44.01}, "N2": {"Mw": 28.01},
}


class PengRobinsonEOS:
    """From-scratch PR EOS with binary interaction parameters (kij)."""

    def __init__(self, component_names):
        self.names = list(component_names)
        self.nc = len(self.names)

        # Critical properties, Cp_ig, and Hvap from thermo database
        thermo_names = [_NAME_MAP[n] for n in self.names]
        constants, props = ChemicalConstantsPackage.from_IDs(thermo_names)
        self.Tc = np.array(constants.Tcs)
        self.Pc = np.array(constants.Pcs)
        self.omega = np.array(constants.omegas)
        self.Mw = np.array([COMPONENT_DB[n]["Mw"] for n in self.names])
        self._Cp_ig = props.HeatCapacityGases
        self._Hvap = props.EnthalpyVaporizations

        # kij submatrix for our components
        full_kij = build_kij_matrix()
        self.kij = np.zeros((self.nc, self.nc))
        for i, ni in enumerate(self.names):
            for j, nj in enumerate(self.names):
                self.kij[i, j] = full_kij[COMP_INDEX[ni], COMP_INDEX[nj]]

        # PR pure-component constants
        self.ai0 = 0.45724 * R**2 * self.Tc**2 / self.Pc
        self.bi = 0.07780 * R * self.Tc / self.Pc
        self.kappa = 0.37464 + 1.54226 * self.omega - 0.26992 * self.omega**2

    # ================================================================
    # Alpha function and derivative
    # ================================================================
    def _alpha(self, T):
        """alpha_i(T) = [1 + kappa*(1 - sqrt(T/Tc))]^2"""
        f = 1.0 + self.kappa * (1.0 - np.sqrt(max(T, 1.0) / self.Tc))
        return f * f

    def _dalpha_dT(self, T):
        """d(alpha_i)/dT"""
        T = max(T, 1.0)
        sqrt_Tr = np.sqrt(T / self.Tc)
        f = 1.0 + self.kappa * (1.0 - sqrt_Tr)
        return -self.kappa * f / np.sqrt(T * self.Tc)

    # ================================================================
    # Mixing rules with kij
    # ================================================================
    def _mix_params(self, z, T):
        """
        Compute a_m, b_m, da_m/dT with van der Waals one-fluid mixing rules.
        Returns (am, bm, dam_dT, ai_T, aij).
        """
        z = np.asarray(z, dtype=float)
        z = np.maximum(z, 1e-30)
        z = z / np.sum(z)

        alpha = self._alpha(T)
        ai_T = self.ai0 * alpha
        sqrt_ai = np.sqrt(np.maximum(ai_T, 0.0))

        # a_ij = sqrt(ai*aj) * (1-kij)
        aij = np.outer(sqrt_ai, sqrt_ai) * (1.0 - self.kij)

        am = z @ aij @ z
        bm = z @ self.bi

        # da_m/dT for enthalpy departure
        dalpha = self._dalpha_dT(T)
        dai_T = self.ai0 * dalpha

        nc = self.nc
        daij = np.zeros((nc, nc))
        for i in range(nc):
            for j in range(nc):
                sij = sqrt_ai[i] * sqrt_ai[j]
                if sij > 1e-30:
                    daij[i, j] = (1.0 - self.kij[i, j]) * \
                        (dai_T[i] * ai_T[j] + ai_T[i] * dai_T[j]) / (2.0 * sij)
        dam_dT = z @ daij @ z

        return am, bm, dam_dT, ai_T, aij

    # ================================================================
    # Cubic solver
    # ================================================================
    def _solve_cubic(self, A, B):
        """
        Solve PR cubic: Z^3 - (1-B)Z^2 + (A-3B^2-2B)Z - (AB-B^2-B^3) = 0.
        Returns (Z_V, Z_L).

        When only 1 real root exists (supercritical mixture), the real
        part of the complex conjugate pair is used as the "virtual"
        root for the missing phase.
        """
        c2 = -(1.0 - B)
        c1 = A - 3.0 * B * B - 2.0 * B
        c0 = -(A * B - B * B - B * B * B)

        roots = np.roots([1.0, c2, c1, c0])

        real_roots = []
        complex_real_part = None
        for r in roots:
            if abs(r.imag) < 1e-8 and r.real > 0:
                real_roots.append(r.real)
            elif complex_real_part is None and r.real > 0:
                complex_real_part = r.real

        if len(real_roots) >= 2:
            real_roots.sort()
            Z_V = real_roots[-1]
            Z_L = max(real_roots[0], B + 1e-6)
            return Z_V, Z_L

        if len(real_roots) == 1:
            Z_real = real_roots[0]
            if complex_real_part is not None and complex_real_part > B:
                Z_virtual = max(complex_real_part, B + 1e-6)
                if Z_real > Z_virtual:
                    return Z_real, Z_virtual
                else:
                    return Z_virtual, Z_real
            return Z_real, max(B + 1e-6, Z_real * 0.15)

        Z = max(B + 0.01, 0.3)
        return Z, Z

    # ================================================================
    # Fugacity coefficients
    # ================================================================
    def fugacity_coefficients(self, T, P, z, phase='vapor'):
        """
        Compute ln(phi_i) for each component.

        ln(phi_i) = (bi/bm)(Z-1) - ln(Z-B)
                   - A/(2*sqrt(2)*B) * [2*sum_j(zj*aij)/am - bi/bm]
                     * ln[(Z+(1+sqrt2)*B)/(Z+(1-sqrt2)*B)]
        """
        am, bm, _, ai_T, aij = self._mix_params(z, T)

        A = am * P / (R * T) ** 2
        B = bm * P / (R * T)

        Z_V, Z_L = self._solve_cubic(A, B)
        Z = Z_V if phase == 'vapor' else Z_L

        z = np.maximum(np.asarray(z, dtype=float), 1e-30)
        z = z / np.sum(z)

        sum_za = aij @ z  # sum_j z_j*a_ij for each i

        zmB = max(Z - B, 1e-30)
        arg = max(Z + (1.0 + SQRT2) * B, 1e-30) / max(Z + (1.0 - SQRT2) * B, 1e-30)
        ln_ratio = np.log(max(arg, 1e-30))

        fac = A / (2.0 * SQRT2 * max(B, 1e-30))
        am_safe = max(am, 1e-30)
        bm_safe = max(bm, 1e-30)

        ln_phi = np.zeros(self.nc)
        for i in range(self.nc):
            bi_bm = self.bi[i] / bm_safe
            ln_phi[i] = (bi_bm * (Z - 1.0)
                         - np.log(zmB)
                         - fac * (2.0 * sum_za[i] / am_safe - bi_bm) * ln_ratio)

        return ln_phi

    def K_values(self, T, P, x, y):
        """K_i = phi_L_i / phi_V_i from PR fugacity coefficients."""
        ln_phi_L = self.fugacity_coefficients(T, P, x, 'liquid')
        ln_phi_V = self.fugacity_coefficients(T, P, y, 'vapor')
        return np.exp(np.clip(ln_phi_L - ln_phi_V, -50, 50))

    def K_values_wilson(self, T, P):
        """Wilson correlation for initial K estimates."""
        return (self.Pc / P) * np.exp(5.373 * (1.0 + self.omega) * (1.0 - self.Tc / T))

    # ================================================================
    # Enthalpy
    # ================================================================
    def enthalpy_departure(self, T, P, z, phase='vapor'):
        """
        PR departure enthalpy (J/mol).

        H_dep = RT(Z-1) + (T*da/dT - a)/(2*sqrt(2)*b)
                * ln[(Z+(1+sqrt2)*B)/(Z+(1-sqrt2)*B)]
        """
        am, bm, dam_dT, _, _ = self._mix_params(z, T)

        A = am * P / (R * T) ** 2
        B = bm * P / (R * T)

        Z_V, Z_L = self._solve_cubic(A, B)
        Z = Z_V if phase == 'vapor' else Z_L

        arg = max(Z + (1.0 + SQRT2) * B, 1e-30) / max(Z + (1.0 - SQRT2) * B, 1e-30)
        ln_ratio = np.log(max(arg, 1e-30))

        H_dep = R * T * (Z - 1.0)
        if bm > 1e-30:
            H_dep += (T * dam_dT - am) / (2.0 * SQRT2 * bm) * ln_ratio

        return H_dep

    def ideal_gas_enthalpy(self, T, z):
        """Ideal gas enthalpy (J/mol) from Cp_ig integration, T_ref=298.15 K."""
        z = np.maximum(np.asarray(z, dtype=float), 1e-30)
        z = z / np.sum(z)
        T_ref = 298.15
        H_ig = 0.0
        for i in range(self.nc):
            if z[i] > 1e-20:
                try:
                    H_ig += z[i] * self._Cp_ig[i].T_dependent_property_integral(T_ref, T)
                except Exception:
                    H_ig += z[i] * 35.0 * (T - T_ref)  # rough fallback
        return H_ig

    def mixture_enthalpy(self, T, P, z, phase='vapor'):
        """Total mixture enthalpy (J/mol) = H_ig + H_dep (PR EOS)."""
        return self.ideal_gas_enthalpy(T, z) + self.enthalpy_departure(T, P, z, phase)

    def pure_component_enthalpy(self, T, P, z, phase='vapor'):
        """
        Mixture enthalpy from pure component enthalpies (ideal mixing).

        This is the approach used by ProSim, Mixprops, and similar tools:
        - Vapor: H = sum_i z_i * H_ig_i(T)
        - Liquid: h = sum_i z_i * [H_ig_i(T) - dHvap_i(T)]

        For supercritical components (T >= Tc_i), dHvap = 0 so h_i = H_ig_i.
        Cp_ig and dHvap correlations from DIPPR (via thermo library).
        """
        z = np.maximum(np.asarray(z, dtype=float), 1e-30)
        z = z / np.sum(z)
        T_ref = 298.15

        H = 0.0
        for i in range(self.nc):
            if z[i] < 1e-20:
                continue

            # Ideal gas enthalpy: integral(Cp_ig, T_ref, T)
            try:
                H_ig_i = self._Cp_ig[i].T_dependent_property_integral(T_ref, T)
            except Exception:
                H_ig_i = 35.0 * (T - T_ref)

            if phase == 'liquid' and T < self.Tc[i]:
                # Subcritical: subtract latent heat of vaporization
                try:
                    dHvap = self._Hvap[i].T_dependent_property(T)
                    if dHvap is not None and dHvap > 0:
                        H += z[i] * (H_ig_i - dHvap)
                    else:
                        H += z[i] * H_ig_i
                except Exception:
                    H += z[i] * H_ig_i
            else:
                # Vapor, or supercritical liquid component
                H += z[i] * H_ig_i

        return H

    # ================================================================
    # Flash TP (Rachford-Rice + successive substitution)
    # ================================================================
    def flash_TP(self, T, P, z):
        """TP flash. Returns (VF, x, y, K)."""
        z = np.maximum(np.asarray(z, dtype=float), 1e-30)
        z = z / np.sum(z)

        K = self.K_values_wilson(T, P)

        for outer in range(30):
            if np.sum(z * K) <= 1.0:
                return 0.0, z.copy(), z.copy(), K
            if np.sum(z / K) <= 1.0:
                return 1.0, z.copy(), z.copy(), K

            VF = 0.5
            for _ in range(100):
                Km1 = K - 1.0
                d = 1.0 + VF * Km1
                f = np.sum(z * Km1 / d)
                df = -np.sum(z * Km1 ** 2 / d ** 2)
                if abs(df) < 1e-30:
                    break
                dV = -f / df
                VF = np.clip(VF + dV, 1e-10, 1.0 - 1e-10)
                if abs(dV) < 1e-12:
                    break

            d = 1.0 + VF * (K - 1.0)
            x = z / d
            x /= np.sum(x)
            y = K * x
            y /= np.sum(y)

            K_new = self.K_values(T, P, x, y)
            if np.max(np.abs(np.log(np.maximum(K_new, 1e-30) / np.maximum(K, 1e-30)))) < 1e-6:
                K = K_new
                break
            K = 0.5 * K + 0.5 * K_new

        d = 1.0 + VF * (K - 1.0)
        x = z / d
        x /= np.sum(x)
        y = K * x
        y /= np.sum(y)
        return VF, x, y, K

    # ================================================================
    # Bubble point temperature
    # ================================================================
    def bubble_T(self, P, x, T_guess=None):
        """Bubble point T by Newton-Raphson with PR K-values."""
        x = np.maximum(np.asarray(x, dtype=float), 1e-30)
        x = x / np.sum(x)

        T = float(T_guess) if T_guess else np.dot(x, self.Tc) * 0.7
        y = x.copy()

        for _ in range(60):
            K = self.K_values(T, P, x, y)
            y = K * x
            ys = np.sum(y)
            y = y / max(ys, 1e-30)
            f = ys - 1.0

            dT = max(0.1, abs(T) * 1e-4)
            K_p = self.K_values(T + dT, P, x, y)
            f_p = np.sum(K_p * x) - 1.0
            df = (f_p - f) / dT
            if abs(df) < 1e-30:
                break
            delta = np.clip(-f / df, -30, 30)
            T = np.clip(T + delta, 100, 700)
            if abs(delta) < 0.01:
                break

        K = self.K_values(T, P, x, y)
        y = K * x
        y /= max(np.sum(y), 1e-30)
        return T, K, y



from thermochem import burcat, combustion
from ThermodynamicCycles.FluidPort.FluidPort import FluidPort
from CoolProp.CoolProp import PropsSI
import pandas as pd 

class Object:
    def __init__(self):
        self.air_Inlet=FluidPort() 
        self.air_Inlet.fluid="air"
        self.Inlet=FluidPort()
        self.Inlet.fluid="water"
        self.Outlet=FluidPort()
        self.Outlet.fluid="water"

    def calculate (self):
        #calcul de la température d'air
        self.Ti_air=PropsSI('T','P',self.air_Inlet.P,'H',self.air_Inlet.h,self.air_Inlet.fluid) # K
        print("self.Ti_air",self.Ti_air)
        #print("self.Ti_air=",self.Ti_air)


from ThermodynamicCycles.FluidPort.FluidPort import FluidPort
from CoolProp.CoolProp import PropsSI
import pandas as pd 

import matplotlib.pyplot as plt
import numpy as np
import random
from scipy.optimize import minimize
from matplotlib.ticker import FuncFormatter

class Object:
    def __init__(self):
        self.Timestamp=None

        #Input and Output Connector
        self.Inlet=FluidPort() 
        self.Outlet=FluidPort()

        # #Input Data
        self.eta=None
        self.Pdischarge_bar=None
        # self.Tcond_degC=None
        self.Pdischarge=None #self.Pdischarge_bar*100000
        # self.Tdischarge_target=None #°C

        #points de fonctionnement de la pompe
        self.X_F = None #débit volumique m3/h X_F = [5,34,40,50]
      
        self.Y_hmt = None # Hauteur manométrique Y_hmt = [12,60,80,68]
        self.Y_eta = None # point de rendement Y_eta = [0.4,0.8,0.9,0.1]

        # Modèles et données pour la courbe
        self.model_hmt = None
        self.model_eta = None
        self.polynomial_features = None
        self.nb_degree = None
        self.x_new_min = None
        self.x_new_max = None
        self.Y_hmt_NEW = None
        self.Y_eta_NEW = None

        # #Initial Values
        #self.Inlet.fluid=None
        # self.Inlet.P=101325
        # self.F=0.1
        # self.Inlet.F=self.F

        # self.F_Sm3s=None
        # self.F_Sm3h=None
        self.F_Sm3s=None
        
        #Output Data
        self.df=[]

        self.Q_pump=0
    #     self.Q_losses=0
        self.Ti_degC=None
        
        # Callback pour détecter les changements de pression en aval
        self.Outlet.callback = self.on_pressure_change
        
    def on_pressure_change(self):
        """
        Callback appelé lorsque self.Outlet.P change (pression imposée par le réseau en aval).
        Recalcule le débit de la pompe en fonction de la nouvelle pression de refoulement.
        """
        if hasattr(self, '_calculating') or hasattr(self, '_calculating_inverse'):
            return  # Éviter les boucles infinies pendant le calcul
        
        print(f"🔄 PUMP CALLBACK: Détection changement pression Outlet: {self.Outlet.P/100000:.3f} bar")
        
        # Si la pression de sortie change, recalculer le point de fonctionnement
        if self.Outlet.P is not None and self.Inlet.P is not None:
            self._calculating_inverse = True
            try:
                self.calculate_from_network_pressure()
            finally:
                delattr(self, '_calculating_inverse')
    
    def calculate_from_network_pressure(self):
        """
        Recalcule le débit de la pompe en fonction de la pression imposée par le réseau.
        Appelée automatiquement via callback quand Outlet.P change.
        """
        if self.model_hmt is None:
            print("⚠️  Modèle HMT non initialisé. Impossible de recalculer le débit.")
            return
        
        # Calculer le nouveau delta_P imposé par le réseau
        self.delta_p = self.Outlet.P - self.Inlet.P
        
        # Calculer la HMT correspondante
        rho = PropsSI("D", "P", self.Inlet.P, "T", self.Ti_degC + 273.15, self.Inlet.fluid)
        g = 9.81
        self.hmt = self.delta_p / (rho * g)
        
        print(f"   ΔP réseau: {self.delta_p/100000:.3f} bar → HMT: {self.hmt:.2f} m")
        
        # Trouver le débit correspondant sur la courbe caractéristique
        def error_function(F_m3h):
            F_m3h_array = np.array(F_m3h).reshape(-1, 1)
            F_transformed = self.polynomial_features.transform(F_m3h_array)
            hmt_predicted = self.model_hmt.predict(F_transformed)[0][0]
            return abs(hmt_predicted - self.hmt)
        
        # Recherche du débit
        initial_guess = self.F_m3h if self.F_m3h is not None else self.X_F[0]
        bounds = [(0.1, 1.5 * max(self.X_F))]
        
        result = minimize(
            error_function,
            x0=initial_guess,
            bounds=bounds,
            method='L-BFGS-B',
            options={'ftol': 1e-6, 'disp': False}
        )
        
        if result.success:
            old_F_m3h = self.F_m3h
            self.F_m3h = result.x[0]
            
            # Recalculer le débit massique
            self.F_m3s = self.F_m3h / 3600
            self.Inlet.F = self.F_m3s * rho
            self.Outlet.F = self.Inlet.F
            
            # Recalculer le rendement
            self.calculate_eta()
            
            print(f"   ✓ Débit adapté: {old_F_m3h:.2f} → {self.F_m3h:.2f} m³/h (F={self.Inlet.F:.4f} kg/s)")
            print(f"   η = {self.eta*100:.1f}%")
            
            # Propager le changement en amont si nécessaire
            if hasattr(self.Inlet, 'callback') and self.Inlet.callback:
                self.Inlet.callback()
        else:
            print(f"⚠️  Échec recherche débit: {result.message}")
        
    def calculate (self):
        # Marquer qu'on est en train de calculer pour éviter le callback
        self._calculating = True
        
        if self.Pdischarge_bar is not None:
            self.Pdischarge=self.Pdischarge_bar*100000

        self.Ti_degC=-273.15+PropsSI("T", "P", self.Inlet.P, "H", self.Inlet.h, self.Inlet.fluid)
        #print("Ti_degC",self.Ti_degC)
        if self.Inlet.F is not None:
            self.F_m3s =self.Inlet.F/PropsSI("D", "P", self.Inlet.P, "T", (self.Ti_degC+273.15), self.Inlet.fluid)
            self.F_m3h=self.F_m3s*3600
        #print("F_m3s",self.F_m3s)

        

        
    #     # outlet connector calculation
        self.Outlet.fluid=self.Inlet.fluid
        self.Outlet.h=self.Inlet.h
        self.Outlet.F=self.Inlet.F
        
        # Désactiver temporairement le callback pour éviter les boucles infinies
        original_callback = self.Outlet.callback
        self.Outlet.callback = None
        
        self.Outlet.P=self.Inlet.P
        
        # Réactiver le callback après l'initialisation
        self.Outlet.callback = original_callback


        #Corrélation de la courbe caractéristique de la pompe
        #pip install scikit-learn
        from sklearn.linear_model import LinearRegression  
        from sklearn.preprocessing import PolynomialFeatures 
        from sklearn.metrics import mean_squared_error, r2_score



        #----------------------------------------------------------------------------------------#
        # Step 1: training data
        if self.X_F is None:
            self.X_F = [7,50,100,150]
        if self.Y_hmt is None:
            self.Y_hmt = [12,60,80,68]
        if self.Y_eta is None:
            self.Y_eta = [0.5,0.7,0.5,0.4]
        
       
        #print(max(self.X_F))
        self.X_F = np.asarray(self.X_F)
        max_x=max(self.X_F)
        self.Y_hmt = np.asarray(self.Y_hmt)
        self.max_Y_hmt=max(self.Y_hmt)
        self.Y_eta = np.asarray(self.Y_eta)
        max_Y_eta=max(self.Y_eta)


        self.X_F = self.X_F[:,np.newaxis]
        self.Y_hmt = self.Y_hmt[:,np.newaxis]
        self.Y_eta = self.Y_eta[:,np.newaxis]


        #----------------------------------------------------------------------------------------#
        # Step 2: data preparation

        self.nb_degree = len(self.X_F)-1

        self.polynomial_features = PolynomialFeatures(degree = self.nb_degree)

        X_TRANSF = self.polynomial_features.fit_transform(self.X_F)
       

        #----------------------------------------------------------------------------------------#
        # Step 3: define and train a model

        self.model_hmt = LinearRegression()
        self.model_eta = LinearRegression()

        #self.model_hmt.fit(X_TRANSF, self.Y_hmt)
        self.model_hmt.fit(X_TRANSF, self.Y_hmt)
        self.model_eta.fit(X_TRANSF, self.Y_eta)

        #----------------------------------------------------------------------------------------#
        # Step 4: calculate bias and variance

        self.Y_hmt_NEW = self.model_hmt.predict(X_TRANSF)
        self.Y_eta_NEW = self.model_eta.predict(X_TRANSF)

        self.rmse_hmt = np.sqrt(mean_squared_error(self.Y_hmt,self.Y_hmt_NEW))
        self.r2_hmt = r2_score(self.Y_hmt,self.Y_hmt_NEW)
        #print('RMSE: ', self.rmse_hmt)
        #print('R2: ', self.r2_hmt)

        rmse_eta = np.sqrt(mean_squared_error(self.Y_eta,self.Y_eta_NEW))
        r2_eta = r2_score(self.Y_eta,self.Y_eta_NEW)
        #print('RMSE: ', rmse_eta)
        #print('R2: ', r2_eta)

        #----------------------------------------------------------------------------------------#
        # Step 5: prediction

        self.x_new_min = 0.0
        # Étendre la plage pour inclure les points de fonctionnement potentiels
        base_max = 1.05 * max_x
        if self.F_m3h is not None:
            # S'assurer que le point de fonctionnement est dans la plage
            self.x_new_max = max(base_max, 1.2 * self.F_m3h)
        else:
            self.x_new_max = base_max

        self.X_NEW = np.linspace(self.x_new_min, self.x_new_max, 100)
        self.X_NEW = self.X_NEW[:,np.newaxis]
        #print(self.X_NEW)

        X_NEW_TRANSF = self.polynomial_features.fit_transform(self.X_NEW)
   

        self.Y_hmt_NEW = self.model_hmt.predict(X_NEW_TRANSF)
        self.Y_eta_NEW = self.model_eta.predict(X_NEW_TRANSF)

        #calculer hmt et delta_p pour le point de fonctionnement self.m3/h
        if self.F_m3h is not None and self.Pdischarge is None:
            self.calculate_hmt()

        if self.Pdischarge is not None:
            self.calculate_flow_rate()
        
        if self.F_m3h is not None:
            self.calculate_eta()

        try:
            self.Q_pump=self.F_m3s*(self.Pdischarge-self.Inlet.P)/self.eta
        except:
            self.Q_pump=self.F_m3s*(self.delta_p)/self.eta

        # Désactiver temporairement le callback pour mettre à jour Outlet.P
        original_callback = self.Outlet.callback
        self.Outlet.callback = None
        
        # Mettre à jour la pression de sortie
        if self.Pdischarge is not None:
            self.Outlet.P = self.Pdischarge
        else:
            self.Outlet.P = self.Inlet.P + self.delta_p
        
        # Réactiver le callback
        self.Outlet.callback = original_callback
        
        # Fin du calcul initial, autoriser les callbacks
        if hasattr(self, '_calculating'):
            delattr(self, '_calculating')

        #-----------------------------------------------------------------------------------------#
        # sortir les résultats sous forme de dataframe avec le débit volumique et la hauteur manométrique du point de fonctionnement

        self.df = pd.DataFrame({'Pump': [self.Timestamp,self.Inlet.fluid,self.Inlet.F,self.F_m3h,self.hmt,self.delta_p,self.Q_pump/1000,self.eta,], },
                      index = ['Timestamp','pump_fluid','pump_F_kgs','pump_F_m3h','hmt(m)','delta_p (Pa)','Qpump(KW)','self.eta' ])


    def calculate_eta(self):
        """
        Calcule le rendement (eta) de la pompe en fonction du débit volumique (F_m3h)
        en utilisant la corrélation.

        Returns:
            float: Rendement (eta) de la pompe.
        """
        if self.model_eta is None:
            raise ValueError("Le modèle de rendement (eta) doit être entraîné avant de l'utiliser.")

        if self.F_m3h is None:
            raise ValueError("Le débit volumique (F_m3h) doit être calculé avant de prédire le rendement.")

        # Transformer le débit volumique pour le modèle
        F_m3h_array = np.array([[self.F_m3h]])  # Assurez-vous que F_m3h est un tableau 2D
        F_transformed = self.polynomial_features.transform(F_m3h_array)

        # Prédire le rendement
        self.eta = self.model_eta.predict(F_transformed)[0][0]
        return self.eta

    def calculate_flow_rate(self):
        """
        Calcule le débit volumique (F_m3h) en fonction des pressions d'entrée et de sortie,
        et de la hauteur manométrique calculée à partir de la corrélation.

        Returns:
            float: Débit volumique (F_m3h) en m³/h.
        """
        if self.Inlet.P is None or self.Pdischarge is None:
            raise ValueError("Les pressions d'entrée et de sortie doivent être définies.")

        if self.model_hmt is None:
            raise ValueError("Le modèle de hauteur manométrique (Hmt) doit être entraîné avant de l'utiliser.")

        # Calculer la hauteur manométrique (Hmt) à partir des pressions
        rho = PropsSI("D", "P", self.Inlet.P, "T", self.Ti_degC + 273.15, self.Inlet.fluid)
        g = 9.81  # Accélération gravitationnelle en m/s²
        self.delta_p = self.Pdischarge - self.Inlet.P
        self.hmt = self.delta_p / (rho * g)
        # print("Hauteur manométrique calculée (self.hmt):", self.hmt)

        # Fonction d'erreur pour minimisation
        def error_function(F_m3h):
            """
            Fonction d'erreur pour la minimisation.
            Compare la hauteur manométrique prédite avec la hauteur calculée.
            """
            F_m3h_array = np.array(F_m3h).reshape(-1, 1)  # Convertir en tableau 2D
            F_transformed = self.polynomial_features.transform(F_m3h_array)
            hmt_predicted = self.model_hmt.predict(F_transformed)[0][0]
            return abs(hmt_predicted - self.hmt)

        # Estimation initiale et bornes
        initial_guess = self.F_m3h if self.F_m3h is not None else self.X_F[0][0]  # Utiliser une estimation existante ou le premier point
        bounds = [(0.1, 1.5 * max(self.X_F))]  # Débit volumique entre 0.1 et 1.5 fois le maximum des données

        # Recherche numérique pour minimiser l'erreur
        result = minimize(
            error_function,
            x0=initial_guess,
            bounds=bounds,
            method='L-BFGS-B',  # Méthode robuste pour les problèmes avec bornes
            options={'ftol': 1e-6, 'disp': False}  # Tolérance plus stricte pour une meilleure précision
        )

        if result.success:
            self.F_m3h = result.x[0]
            # print("Débit volumique calculé (self.F_m3h):", self.F_m3h)
        else:
            raise ValueError(f"La recherche numérique pour le débit volumique a échoué : {result.message}")


    def calculate_hmt(self):
        """
        Calcule la hauteur manométrique (Hmt) à partir de la corrélation pour X_F = self.F_m3h.

        Returns:
            float: Hauteur manométrique prédite (Hmt) en mètres.
        """
        if self.model_hmt is None:
            raise ValueError("Le modèle de hauteur manométrique (Hmt) doit être entraîné avant de l'utiliser.")

        if self.F_m3h is None:
            raise ValueError("Le débit volumique (F_m3h) doit être calculé avant de prédire Hmt.")

        # Transformer le débit volumique pour le modèle
        F_m3h_array = np.array([[self.F_m3h]])
        F_transformed = self.polynomial_features.transform(F_m3h_array)

        # Prédire la hauteur manométrique
        self.hmt = self.model_hmt.predict(F_transformed)[0][0]
        
            # Obtenir la densité du fluide à l'entrée
        rho = PropsSI("D", "P", self.Inlet.P, "T", self.Ti_degC + 273.15, self.Inlet.fluid)

        # Accélération gravitationnelle
        g = 9.81  # m/s²

        # Calcul de la différence de pression
        self.delta_p = rho * g * self.hmt

    

        #----------------------------------------------------------------------------------------#
        # Step 6: Plotting

        
    def plot_pump_curve(self, figsize=(14, 10)):
        """
        Trace la courbe caractéristique de la pompe avec deux graphiques séparés :
        - Graphique supérieur : Hauteur manométrique (Hmt) et pression (ΔP)
        - Graphique inférieur : Rendement (eta)
        Affiche les points de construction, le point de fonctionnement avec correspondance entre les deux graphiques.
        """
        # Calculer la densité du fluide pour convertir Hmt en bar
        rho = PropsSI("D", "P", self.Inlet.P, "T", self.Ti_degC + 273.15, self.Inlet.fluid)
        g = 9.81  # Accélération gravitationnelle en m/s²
        
        # Convertir les données en pression (bar)
        Y_pressure_NEW = self.Y_hmt_NEW * rho * g / 100000  # Conversion Hmt → bar
        Y_pressure_points = self.Y_hmt.flatten() * rho * g / 100000
        
        # Créer une figure avec deux sous-graphiques
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize, sharex=True)
        fig.suptitle('COURBE CARACTÉRISTIQUE DE LA POMPE', fontsize=16, fontweight='bold')
        
        # ========================================================================
        # GRAPHIQUE 1 : HAUTEUR MANOMÉTRIQUE ET PRESSION
        # ========================================================================
        
        # Courbe HMT
        line_hmt = ax1.plot(self.X_NEW, self.Y_hmt_NEW, color='#FF6B35', linewidth=3, 
                            label='Courbe HMT (corrélation)', zorder=5)
        ax1.set_ylabel('Hauteur Manométrique (m)', color='#FF6B35', fontsize=12, fontweight='bold')
        ax1.tick_params(axis='y', labelcolor='#FF6B35', labelsize=10)
        ax1.grid(True, alpha=0.3, linestyle='--')
        
        # Axe secondaire pour la pression
        ax1_pressure = ax1.twinx()
        ax1_pressure.plot(self.X_NEW, Y_pressure_NEW, color='#004E89', linewidth=3, 
                         linestyle='--', label='Courbe ΔP (bar)', alpha=0.7, zorder=5)
        ax1_pressure.set_ylabel('Pression ΔP (bar)', color='#004E89', fontsize=12, fontweight='bold')
        ax1_pressure.tick_params(axis='y', labelcolor='#004E89', labelsize=10)
        
        # Points de construction caractéristiques (HMT)
        ax1.scatter(self.X_F.flatten(), self.Y_hmt.flatten(), color='#FF6B35', 
                   s=150, marker='s', edgecolors='black', linewidth=2, 
                   label='Points construction HMT', zorder=10)
        
        # Annotations pour les points de construction (HMT)
        for i, (x, y) in enumerate(zip(self.X_F.flatten(), self.Y_hmt.flatten())):
            ax1.annotate(f'P{i+1}\n{x:.1f} m³/h\n{y:.1f} m', 
                        xy=(x, y), xytext=(10, 10), textcoords='offset points',
                        fontsize=8, bbox=dict(boxstyle='round,pad=0.5', facecolor='#FF6B35', alpha=0.3),
                        arrowprops=dict(arrowstyle='->', color='#FF6B35', lw=1.5))
        
        # Point de fonctionnement (HMT)
        if self.F_m3h is not None and self.hmt is not None:
            pressure_operating = self.hmt * rho * g / 100000
            
            # Point sur HMT
            ax1.scatter([self.F_m3h], [self.hmt], color='red', s=300, marker='*', 
                       edgecolors='darkred', linewidth=2, label='Point fonctionnement', zorder=15)
            
            # Point sur pression
            ax1_pressure.scatter([self.F_m3h], [pressure_operating], color='red', 
                                s=200, marker='*', edgecolors='darkred', linewidth=2, zorder=15)
            
            # Lignes de correspondance
            ax1.axvline(self.F_m3h, color='red', linestyle=':', linewidth=2, alpha=0.5, zorder=3)
            ax1.axhline(self.hmt, color='red', linestyle=':', linewidth=2, alpha=0.5, zorder=3)
            
            # Annotation du point de fonctionnement
            ax1.annotate(f'FONCTIONNEMENT\nQ = {self.F_m3h:.2f} m³/h\nHMT = {self.hmt:.2f} m\nΔP = {pressure_operating:.3f} bar',
                        xy=(self.F_m3h, self.hmt), xytext=(30, -30), textcoords='offset points',
                        fontsize=10, fontweight='bold',
                        bbox=dict(boxstyle='round,pad=0.8', facecolor='red', alpha=0.2, edgecolor='darkred', linewidth=2),
                        arrowprops=dict(arrowstyle='->', color='red', lw=2))
        
        # Limites et légendes
        ax1.set_ylim(0, 1.1 * self.max_Y_hmt)
        ax1_pressure.set_ylim(0, 1.1 * max(Y_pressure_points))
        
        # Légendes combinées
        lines1, labels1 = ax1.get_legend_handles_labels()
        lines2, labels2 = ax1_pressure.get_legend_handles_labels()
        ax1.legend(lines1 + lines2, labels1 + labels2, loc='upper right', fontsize=9, framealpha=0.9)
        
        # ========================================================================
        # GRAPHIQUE 2 : RENDEMENT
        # ========================================================================
        
        # Courbe rendement
        ax2.plot(self.X_NEW, self.Y_eta_NEW * 100, color='#1B9C85', linewidth=3, 
                label='Courbe rendement (corrélation)', zorder=5)
        ax2.set_xlabel('Débit volumique (m³/h)', fontsize=12, fontweight='bold')
        ax2.set_ylabel('Rendement η (%)', color='#1B9C85', fontsize=12, fontweight='bold')
        ax2.tick_params(axis='y', labelcolor='#1B9C85', labelsize=10)
        ax2.tick_params(axis='x', labelsize=10)
        ax2.grid(True, alpha=0.3, linestyle='--')
        
        # Points de construction caractéristiques (rendement)
        ax2.scatter(self.X_F.flatten(), self.Y_eta.flatten() * 100, color='#1B9C85', 
                   s=150, marker='s', edgecolors='black', linewidth=2,
                   label='Points construction η', zorder=10)
        
        # Annotations pour les points de construction (rendement)
        for i, (x, y) in enumerate(zip(self.X_F.flatten(), self.Y_eta.flatten())):
            ax2.annotate(f'P{i+1}\n{x:.1f} m³/h\n{y*100:.1f}%', 
                        xy=(x, y*100), xytext=(10, -15), textcoords='offset points',
                        fontsize=8, bbox=dict(boxstyle='round,pad=0.5', facecolor='#1B9C85', alpha=0.3),
                        arrowprops=dict(arrowstyle='->', color='#1B9C85', lw=1.5))
        
        # Point de fonctionnement (rendement)
        if self.F_m3h is not None and self.eta is not None:
            ax2.scatter([self.F_m3h], [self.eta * 100], color='red', s=300, marker='*',
                       edgecolors='darkred', linewidth=2, label='Point fonctionnement', zorder=15)
            
            # Lignes de correspondance
            ax2.axvline(self.F_m3h, color='red', linestyle=':', linewidth=2, alpha=0.5, zorder=3)
            ax2.axhline(self.eta * 100, color='red', linestyle=':', linewidth=2, alpha=0.5, zorder=3)
            
            # Annotation du point de fonctionnement
            ax2.annotate(f'FONCTIONNEMENT\nQ = {self.F_m3h:.2f} m³/h\nη = {self.eta*100:.2f}%',
                        xy=(self.F_m3h, self.eta * 100), xytext=(30, 15), textcoords='offset points',
                        fontsize=10, fontweight='bold',
                        bbox=dict(boxstyle='round,pad=0.8', facecolor='red', alpha=0.2, edgecolor='darkred', linewidth=2),
                        arrowprops=dict(arrowstyle='->', color='red', lw=2))
        
        # Limites et légende - S'assurer que le point de fonctionnement est visible
        x_max_display = self.x_new_max
        if self.F_m3h is not None:
            x_max_display = max(x_max_display, self.F_m3h * 1.1)
        
        ax2.set_xlim(self.x_new_min, x_max_display)
        ax2.set_ylim(0, 105)
        ax2.legend(loc='upper right', fontsize=9, framealpha=0.9)
        
        # ========================================================================
        # INFORMATIONS SUPPLÉMENTAIRES
        # ========================================================================
        
        # Ajouter un encadré d'informations
        info_text = f'Corrélation polynomiale : degré {self.nb_degree}\n'
        info_text += f'RMSE (HMT) : {self.rmse_hmt:.4f} m\n'
        info_text += f'R² (HMT) : {self.r2_hmt:.4f}\n'
        info_text += f'Fluide : {self.Inlet.fluid} @ {self.Ti_degC:.1f}°C'
        
        fig.text(0.02, 0.02, info_text, fontsize=9, 
                bbox=dict(boxstyle='round', facecolor='wheat', alpha=0.5),
                verticalalignment='bottom', horizontalalignment='left')
        
        # Ajuster l'espacement
        plt.tight_layout(rect=[0, 0.05, 1, 0.97])
        plt.show() 


import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
from collections import defaultdict
import matplotlib.cm as cm
import matplotlib.colors as mcolors
import matplotlib.patches as patches

def draw_hen_from_df(df):
    G = nx.DiGraph()
    pos = {}
    node_labels = {}

    # Regrouper les échangeurs par flux chaud (HS_name)
    hot_streams = defaultdict(list)
    for i, row in df.iterrows():
        hot_streams[row['HS_name']].append((i, row))

    # Pour retrouver les noeuds par flux froid et température
    cold_in_nodes = defaultdict(lambda: defaultdict(list))   # {cold_name: {temp: [node_id,...]}}
    cold_out_nodes = defaultdict(lambda: defaultdict(list))

    # Position base
    y_base = 0
    y_spacing = 5
    x_spacing = 4

    for j, (hot_name, exchangers) in enumerate(hot_streams.items()):
        y = y_base - j * y_spacing
        x = 0

        # Nœud entrée flux chaud
        hot_in_node = f"{hot_name}_Tin"
        G.add_node(hot_in_node)
        node_labels[hot_in_node] = f"{hot_name}_Tin\n{exchangers[0][1]['HS_Ti']}°C"
        pos[hot_in_node] = (x, y)
        x += x_spacing

        previous_node = hot_in_node

        for exch_index, row in exchangers:
            exch = f"E{exch_index+1}"

            # Décalage horizontal pour une ligne sur deux
            x_offset = 1  # Décalage horizontal pour les lignes impaires

            # Nœuds échangeur
            hot_out_node = f"{row['HS_name']}_Tout_{exch_index}"
            cold_in_node = f"{row['CS_name']}_Tin_{exch_index}"
            cold_out_node = f"{row['CS_name']}_Tout_{exch_index}"

            # Ajouter les nœuds
            G.add_node(exch)
            G.add_node(hot_out_node)
            G.add_node(cold_in_node)
            G.add_node(cold_out_node)

            # Labels
            node_labels[exch] = f"{exch}\nQ={round(row['HeatExchanged'], 1)} kW"  # Arrondi à 1 chiffre après la virgule
            node_labels[hot_out_node] = f"{row['HS_name']}_Tout\n{round(row['HS_To'], 1)}°C"
            node_labels[cold_in_node] = f"{row['CS_name']}_Tin\n{round(row['CS_Ti'], 1)}°C"
            node_labels[cold_out_node] = f"{row['CS_name']}_Tout\n{round(row['CS_To'], 1)}°C"

            # Positions avec décalage horizontal pour une ligne sur deux
            pos[exch] = (x + x_offset + (1.0 if j % 2 == 0 else 0), y)
            pos[hot_out_node] = (x + x_offset + (1.5 if j % 2 == 0 else 0.5) + 1.5, y)
            pos[cold_in_node] = (x + x_offset + (1.0 if j % 2 == 0 else 0), y - 1.5)
            pos[cold_out_node] = (x + x_offset + (1.0 if j % 2 == 0 else 0), y + 1.5)


            # Connexions principales
            G.add_edge(previous_node, exch)
            G.add_edge(exch, hot_out_node)
            G.add_edge(cold_in_node, exch)
            G.add_edge(exch, cold_out_node)

            previous_node = hot_out_node
            x += x_spacing

            # Stocker pour liens froids
            cold_in_nodes[row['CS_name']][row['CS_Ti']].append(cold_in_node)
            cold_out_nodes[row['CS_name']][row['CS_To']].append(cold_out_node)

    # Arrondir les températures pour éviter les problèmes de précision
    def round_temp(temp):
        return round(temp, 1)  # Arrondi à 1 chiffre après la virgule

    # Mise à jour des nœuds et connexions avec températures arrondies
    for j, (hot_name, exchangers) in enumerate(hot_streams.items()):
        for exch_index, row in exchangers:
            # Arrondir les températures
            row['HS_Ti'] = round_temp(row['HS_Ti'])
            row['HS_To'] = round_temp(row['HS_To'])
            row['CS_Ti'] = round_temp(row['CS_Ti'])
            row['CS_To'] = round_temp(row['CS_To'])

            # Mise à jour des nœuds de température
            hot_out_node = f"{row['HS_name']}_Tout_{exch_index}"
            cold_in_node = f"{row['CS_name']}_Tin_{exch_index}"
            cold_out_node = f"{row['CS_name']}_Tout_{exch_index}"

            node_labels[hot_out_node] = f"{row['HS_name']}_Tout\n{row['HS_To']}°C"
            node_labels[cold_in_node] = f"{row['CS_name']}_Tin\n{row['CS_Ti']}°C"
            node_labels[cold_out_node] = f"{row['CS_name']}_Tout\n{row['CS_To']}°C"

            # Stocker les nœuds pour connexions froides
            cold_in_nodes[row['CS_name']][row['CS_Ti']].append(cold_in_node)
            cold_out_nodes[row['CS_name']][row['CS_To']].append(cold_out_node)

    # Connexion flux froids si même flux ET même température sortie-entrée
    for cold_name in cold_out_nodes:
        for temp, out_nodes in cold_out_nodes[cold_name].items():
            in_nodes = cold_in_nodes[cold_name].get(temp, [])
            for out_node in out_nodes:
                for in_node in in_nodes:
                    G.add_edge(out_node, in_node)

    # Identifier les températures extrêmes pour chaque flux
    hot_stream_extremes = {}
    cold_stream_extremes = {}

    for hot_name, exchangers in hot_streams.items():
        temps = [row['HS_Ti'] for _, row in exchangers] + [row['HS_To'] for _, row in exchangers]
        hot_stream_extremes[hot_name] = max(temps)  # Température la plus haute pour le flux chaud

    for cold_name, temp_nodes in cold_in_nodes.items():
        temps = list(temp_nodes.keys()) + list(cold_out_nodes[cold_name].keys())
        cold_stream_extremes[cold_name] = min(temps)  # Température la plus basse pour le flux froid

    # Mise à jour des labels
    for j, (hot_name, exchangers) in enumerate(hot_streams.items()):
        for exch_index, row in exchangers:
            hot_out_node = f"{row['HS_name']}_Tout_{exch_index}"
            cold_in_node = f"{row['CS_name']}_Tin_{exch_index}"
            cold_out_node = f"{row['CS_name']}_Tout_{exch_index}"

            # Mise à jour des labels pour les nœuds de température
            if row['HS_To'] == hot_stream_extremes[row['HS_name']]:
                node_labels[hot_out_node] = f"{row['HS_name']}_Tout\n{round(row['HS_To'], 1)}°C"
            else:
                node_labels[hot_out_node] = f"{round(row['HS_To'], 1)}°C"

            if row['CS_Ti'] == cold_stream_extremes[row['CS_name']]:
                node_labels[cold_in_node] = f"{row['CS_name']}_Tin\n{round(row['CS_Ti'], 1)}°C"
            else:
                node_labels[cold_in_node] = f"{round(row['CS_Ti'], 1)}°C"

            if row['CS_To'] == cold_stream_extremes[row['CS_name']]:
                node_labels[cold_out_node] = f"{row['CS_name']}_Tout\n{round(row['CS_To'], 1)}°C"
            else:
                node_labels[cold_out_node] = f"{round(row['CS_To'], 1)}°C"

    # Séparer les nœuds par type avec une vérification stricte
    exchanger_nodes = [node for node in G.nodes if node.startswith("E") and node[1:].isdigit()]
    temperature_nodes = [node for node in G.nodes if node not in exchanger_nodes]

    # Générer une palette de couleurs suffisamment grande pour les flux froids
    num_cold_streams = len(cold_in_nodes)
    distinct_colors = cm.get_cmap('tab20', num_cold_streams)  # Palette avec 20 couleurs distinctes

    # Assigner une couleur unique à chaque flux chaud
    hot_stream_color_map = {hot_name: mcolors.to_hex(cm.get_cmap('Reds', len(hot_streams))(i)) for i, hot_name in enumerate(hot_streams)}

    # Assigner une couleur unique à chaque flux froid (sans chevauchement)
    cold_stream_color_map = {
        cold_name: mcolors.to_hex(distinct_colors(i))  # Utilisation de couleurs uniques
        for i, cold_name in enumerate(cold_in_nodes)
    }

    # Tracé
    plt.figure(figsize=(13.33, 7.5))  # Pour un bon ajustement
    ax = plt.gca()  # Obtenir l'axe pour ajouter des formes personnalisées

    # Couleur fixe pour les flux chauds
    hot_stream_color = 'red'  # Couleur rouge pour tous les flux chauds

    # Dessiner les nœuds de température avec une couleur rouge pour tous les flux chauds
    for hot_name, exchangers in hot_streams.items():
        hot_nodes = [f"{row['HS_name']}_Tout_{exch_index}" for exch_index, row in exchangers]
        hot_nodes.append(f"{hot_name}_Tin")  # Ajouter le nœud d'entrée
     
    # Dessiner les nœuds de température avec des couleurs spécifiques par flux froid
    for i, (cold_name, temp_nodes) in enumerate(cold_in_nodes.items()):
        cold_nodes = []
        for temp_nodes_list in temp_nodes.values():
            cold_nodes.extend(temp_nodes_list)
        for temp_nodes_list in cold_out_nodes[cold_name].values():
            cold_nodes.extend(temp_nodes_list)
      
    # Dessiner les nœuds d'échangeurs (rectangles arrondis)
    for node in exchanger_nodes:
        x, y = pos[node]  # Position du nœud
        label = node_labels[node]  # Texte du nœud
        width = 1.2  # Réduction de la largeur (par exemple, de 1.5 à 1.2)
        height = 0.6  # Réduction de la hauteur (par exemple, de 0.8 à 0.6)
        rect = patches.FancyBboxPatch(
            (x - width / 2, y - height / 2),  # Position du coin inférieur gauche
            width,  # Largeur
            height,  # Hauteur
            boxstyle="round,pad=0.3",  # Style de rectangle arrondi
            edgecolor="red",  # Couleur du bord
            facecolor="green",  # Couleur de remplissage
            linewidth=0.6 # Épaisseur du bord
        )
        ax.add_patch(rect)
        ax.text(x, y, label, ha="center", va="center", fontsize=6, color="black")  # Ajouter le texte au centre

    # Dessiner les nœuds de température (rectangles arrondis uniquement)
    for node in temperature_nodes:
        x, y = pos[node]  # Position du nœud
        label = node_labels[node]  # Texte du nœud
        # Vérifier si la clé existe dans cold_stream_color_map
        node_key = node.split("_")[0]
        color = (
            hot_stream_color
            if node in hot_stream_color_map
            else cold_stream_color_map.get(node_key, "lightblue")  # Utiliser une couleur par défaut si la clé n'existe pas
        )
        width = 0.6  # Largeur du rectangle
        height = 0.25  # Hauteur du rectangle
        rect = patches.FancyBboxPatch(
            (x - width / 2, y - height / 2),  # Position du coin inférieur gauche
            width,  # Largeur
            height,  # Hauteur
            boxstyle="round,pad=0.3",  # Style de rectangle arrondi
            edgecolor="none",  # Couleur du bord
            facecolor=color,  # Couleur de remplissage
            linewidth=1.5,  # Épaisseur du bord
        )
        ax.add_patch(rect)
        ax.text(x, y, label, ha="center", va="center", fontsize=6, color="black")  # Ajouter le texte au centre

    # Dessiner les arêtes
    nx.draw_networkx_edges(G, pos, arrows=True, arrowstyle='->', arrowsize=10)

    # Ajouter les étiquettes pour les nœuds de température
    nx.draw_networkx_labels(G, pos, labels={k: v for k, v in node_labels.items() if k not in exchanger_nodes}, font_size=6)

    # Titre et mise en page
    plt.title("Réseau d'échangeurs de chaleur à partir d'un DataFrame", fontsize=10)
    plt.axis('off')
    plt.tight_layout()
    plt.show()






from OpenWeatherMap import get_weather
import pandas as pd
from datetime import datetime
#import get_weather
#pip install requests[security]

def API_call_location(lat:str,lon:str):

# )
    api_key = get_weather.get_api_key()
#    print(api_key)
    
    location=get_weather.get_location() #je passe à la ligne en dessous pour ajouter une lat et long comme paramètre
#    print(location[0])
    
    list_location=list(location)  #TypeError: 'tuple' object does not support item assignment
    list_location[1]=lat
    list_location[2]=lon
    
    location=tuple(list_location)
    #print("location",location)

    weather = get_weather.get_weather(api_key, location[0],location[1],location[2])
# 
    #print(weather['main']['temp']-273.15,"°C",weather['main']['humidity'],"%")
   # print(weather)
    T=weather['main']['temp']-273.15
    RH=weather['main']['humidity']
    Timestamp=datetime.now()
    df=pd.DataFrame({"Timestamp":Timestamp,"T(°C)":[T],"RH(%)":[RH]})
    return df

#API_T_RH=API_call()
#print(API_T_RH[0])
 
# if __name__ == '__main__':
#     main()


import pandas as pd

def composite_curve(self):
    # Définition des fonctions d'agrégation pour chaque colonne
    agg_functions = {'delta_H': 'sum', 'mCp': 'sum', 'Tsup': 'last', 'Tinf': 'last', 'StreamName': lambda x: list(x)}

    # Groupement par IntervalName et StreamType avec application des fonctions d'agrégation
    composite_curve = self.df_intervals.groupby(['IntervalName', 'StreamType']).agg(agg_functions).reset_index()
    composite_curve = composite_curve.sort_values(by=['StreamType', 'Tsup'], ascending=[True, True])

    # Créer des copies des DataFrames résultants
    self.cold_composite_curve = composite_curve[composite_curve['StreamType'] == 'CS'].copy()
    self.hot_composite_curve = composite_curve[composite_curve['StreamType'] == 'HS'].copy()

    # Calcul de la somme cumulée par StreamType
    self.cold_composite_curve['delta_H'] =-1*self.cold_composite_curve['delta_H']
    self.cold_composite_curve['mCp'] =-1*self.cold_composite_curve['mCp']
    self.cold_composite_curve['cumulative_delta_H'] = self.cold_composite_curve['delta_H'].cumsum()
    self.hot_composite_curve['cumulative_delta_H'] = self.hot_composite_curve['delta_H'].cumsum()

    self.cold_stream=max(self.cold_composite_curve['cumulative_delta_H'])
    self.hot_stream=max(self.hot_composite_curve['cumulative_delta_H'])

    self.heat_recovery=self.hot_stream-self.Cooling_duty
    #self.heat_recovery=self.cold_stream-self.Heating_duty


    # Afficher le résultat
    self.cold_composite_curve
    self.hot_composite_curve


    # Création du nouveau dataframe avec une ligne supplémentaire
    hcc_data = {'T': [self.hot_composite_curve['Tinf'].min()] + self.hot_composite_curve['Tsup'].tolist(),
                'Q': [0] + self.hot_composite_curve['cumulative_delta_H'].tolist()}



    self.df_hcc = pd.DataFrame(hcc_data)
    self.df_hcc


    # Création du nouveau dataframe avec une ligne supplémentaire
    ccc_data = {'T': [self.cold_composite_curve['Tinf'].min()] + self.cold_composite_curve['Tsup'].tolist(),
                'Q': [0] + self.cold_composite_curve['cumulative_delta_H'].tolist()+self.Cooling_duty}

    self.df_ccc = pd.DataFrame(ccc_data)
    self.df_ccc



import CoolProp.CoolProp as CP
import math
import pandas as pd
from AHU.air_humide import air_humide
from scipy.optimize import fsolve  # Add import for solver

class Object:
    DN_TABLE = {
        6: {'di': 0.0078, 'de': 0.0102, 'thickness': 0.0016},
        8: {'di': 0.0109, 'de': 0.0135, 'thickness': 0.0023},
        10: {'di': 0.0136, 'de': 0.0172, 'thickness': 0.0028},
        15: {'di': 0.0169, 'de': 0.0213, 'thickness': 0.0027},
        20: {'di': 0.0217, 'de': 0.0269, 'thickness': 0.0032},
        25: {'di': 0.0273, 'de': 0.0337, 'thickness': 0.0032},
        32: {'di': 0.0358, 'de': 0.0424, 'thickness': 0.0033},
        40: {'di': 0.0411, 'de': 0.0483, 'thickness': 0.0036},
        50: {'di': 0.0525, 'de': 0.0603, 'thickness': 0.0039},
        65: {'di': 0.0683, 'de': 0.0761, 'thickness': 0.0039},
        80: {'di': 0.0809, 'de': 0.0889, 'thickness': 0.0040},
        100: {'di': 0.1061, 'de': 0.1143, 'thickness': 0.0041},
        125: {'di': 0.1311, 'de': 0.1397, 'thickness': 0.0043},
        150: {'di': 0.1593, 'de': 0.1683, 'thickness': 0.0045},
        200: {'di': 0.2091, 'de': 0.2191, 'thickness': 0.0050},
        250: {'di': 0.2618, 'de': 0.2730, 'thickness': 0.0056},
        300: {'di': 0.3123, 'de': 0.3239, 'thickness': 0.0058},
        350: {'di': 0.3434, 'de': 0.3556, 'thickness': 0.0061},
        400: {'di': 0.3934, 'de': 0.4064, 'thickness': 0.0065},
        450: {'di': 0.4432, 'de': 0.4572, 'thickness': 0.0070},
        500: {'di': 0.4930, 'de': 0.5080, 'thickness': 0.0075},
        550: {'di': 0.5428, 'de': 0.5588, 'thickness': 0.0080},
        600: {'di': 0.5926, 'de': 0.6096, 'thickness': 0.0085},
        650: {'di': 0.6424, 'de': 0.6604, 'thickness': 0.0090},
        700: {'di': 0.6922, 'de': 0.7112, 'thickness': 0.0095},
        750: {'di': 0.7420, 'de': 0.7620, 'thickness': 0.0100},
        800: {'di': 0.7918, 'de': 0.8128, 'thickness': 0.0105},
        850: {'di': 0.8416, 'de': 0.8636, 'thickness': 0.0110},
        900: {'di': 0.8914, 'de': 0.9144, 'thickness': 0.0115},
        950: {'di': 0.9412, 'de': 0.9652, 'thickness': 0.0120},
        1000: {'di': 0.9910, 'de': 1.0160, 'thickness': 0.0125},
        1050: {'di': 1.0408, 'de': 1.0668, 'thickness': 0.0130}
    }

    MATERIALS = {
        'Cuivre': 380,
        'Plomb': 35,
        'Acier': 50,
        'Aluminium 99%': 160,
        'Fonte': 50,
        'Zinc': 110
    }

    INSULATION_MATERIALS = {
        'aucun isolant': 0,
        'laine minérale': 0.04,
        'polyuréthanne PUR': 0.03,
        'polystyrène': 0.036,
        'polyéthylène': 0.038,
        'Liège (ICB)': 0.05,
        'Laine minérale (MW)': 0.045,
        'Polystyrène expansé (EPS)': 0.045,
        'Polyéthylène extrudé (PEF)': 0.045,
        'Mousse phénolique – revêtu (PF)': 0.045,
        'Polyuréthane – revêtu (PUR/PIR)': 0.035,
        'Polystyrène extrudé (XPS)': 0.04,
        'Verre cellulaire (CG)': 0.055,
        'Perlite (EPB)': 0.06,
        'Vermiculite': 0.065,
        'Vermiculite expansée (panneaux)': 0.09
    }

    FLUID_PROPERTIES = {
        'Huile thermique': {
            'Cp': 2632.1544,  # J/kg.K
            'mu': 0.1,  # Pa.s
            'k': 0.21,  # W/m.K
            'rho': 900  # kg/m3
        }
    }

    def __init__(self, fluid, T_fluid, F_m3h, DN, material='Acier', insulation='laine minérale', insulation_thickness=0.04, emissivity=0.01, Tamb=20, humidity=40,L_tube=1):
        self.fluid = fluid
        self.Tfluid = T_fluid
        self.F_m3h = F_m3h
        self.Tamb = Tamb
        self.RH = humidity
        self.Tamb = Tamb  # Assuming T1 is the ambient temperature


        # Convert temperatures to Kelvin for calculations
        self.Tamb_K = self.Tamb + 273.15
        self.Tamb_K = self.Tamb + 273.15
        self.Tfluid_K = self.Tfluid + 273.15

        # Pipe properties
        self.dn = DN / 1000  # Convert DN from mm to meters
        self.material = material  # Pipe material
        if DN in self.DN_TABLE:
            self.di = self.DN_TABLE[DN]['di']
            self.de = self.DN_TABLE[DN]['de']
        else:
            raise ValueError(f"DN {DN} not found in the DN_TABLE.")
        self.L_tube = L_tube  # L_tube in meters
        if material in self.MATERIALS:
            self.k_pipe = self.MATERIALS[material]  # Thermal conductivity of the pipe in W/m-°C
        else:
            raise ValueError(f"Material '{material}' not found in the MATERIALS table.")

        # Insulation properties
        self.insulation = insulation
        self.insulation_thickness = insulation_thickness
        self.emissivity = emissivity
        if insulation in self.INSULATION_MATERIALS:
            self.k_insulation = self.INSULATION_MATERIALS[insulation]  # Thermal conductivity of the insulation in W/m-°C
        else:
            raise ValueError(f"Insulation '{insulation}' not found in the INSULATION_MATERIALS table.")

        # Calculate outer diameter with insulation
        self.de_with_insulation = self.de + 2 * self.insulation_thickness

        # Fluid properties
        if fluid in self.FLUID_PROPERTIES:
            self.Cp = self.FLUID_PROPERTIES[fluid]['Cp']
            self.mu = self.FLUID_PROPERTIES[fluid]['mu']
            self.k = self.FLUID_PROPERTIES[fluid]['k']
            self.rho = self.FLUID_PROPERTIES[fluid]['rho']
        else:
            # Using CoolProp to get properties of fluid
            self.Cp = CP.PropsSI('C', 'T', self.Tfluid_K, 'P', 101325, fluid)
            self.mu = CP.PropsSI('V', 'T', self.Tfluid_K, 'P', 101325, fluid)
            self.k = CP.PropsSI('L', 'T', self.Tfluid_K, 'P', 101325, fluid)
            self.rho = CP.PropsSI('D', 'T', self.Tfluid_K, 'P', 101325, fluid)

        # Properties of ambient air
        self.B = 0.0034  # 1/K

        # Using CoolProp to get properties of air
   
        self.mu_air = CP.PropsSI('V', 'T', self.Tamb_K, 'P', 101325, 'Air')
        self.k_air = CP.PropsSI('L', 'T', self.Tamb_K, 'P', 101325, 'Air')
        self.Cp_air = CP.PropsSI('C', 'T', self.Tamb_K, 'P', 101325, 'Air')

        # Calculate Prandtl Number for ambient air
        self.Pr_air = self.Cp_air * self.mu_air / self.k_air
        self.Pr_fluid = self.Cp * self.mu / self.k
   

        # Calculate external surface area of the insulated pipe
        self.A_e_insulated = 3.14159 * self.de_with_insulation * self.L_tube
        self.A_i= 3.14159 * self.di * self.L_tube

        # Initialize DataFrame
        self.df = pd.DataFrame()

        # Initialize Tc and Tf
        self.Tc = self.Tamb + 10.4  # Initial guess for Tc, less than Ta
        self.Tf = (self.Tc + self.Tamb) / 2  # Initial film temperature
        self.rho_air = air_humide.Air_rho_hum(T_db=self.Tf, RH=self.RH, P=101325)

    def calculate_ambient_air_properties(self):
        # Calculate Rayleigh Number
        g = 9.81  # m/s², acceleration due to gravity
        Ra_air = (g * self.B * self.rho_air ** 2 * self.Cp_air * (self.Tc - self.Tamb) * self.de_with_insulation ** 3) / (self.k_air * self.mu_air)
      
        # Calculate Nusselt Number
        Nu = (0.60 + (0.387 * Ra_air ** (1/6)) / (1 + (0.559 / self.Pr_air) ** (9/16)) ** (8/27)) ** 2
      
        
        # Calculate average heat transfer coefficient
        self.h_air = Nu * self.k_air / self.de_with_insulation
        

        return Ra_air, Nu, self.h_air

    def calculate_heat_transfer(self):
        # Convective heat transfer
        q_conv = self.h_air * self.A_e_insulated * (self.Tc - self.Tamb)
        R_conv = (self.Tc - self.Tamb) / q_conv  # Convective resistance

        # Radiative heat transfer
        sigma = 5.670373e-8  # Stefan-Boltzmann constant in W/m²K⁴
        q_rad = self.A_e_insulated * self.emissivity * sigma * ((self.Tc+273.15)**4 - self.Tamb_K**4)
        R_rad = (self.Tc - self.Tamb) / q_rad  # Radiative resistance

        # Calculate internal convective resistance
        R_conv_i = 1 / (self.h_fluid * self.A_i)


        # Calculate conductive resistance of the bare tube : =LN('données d''entrée'!B17/('données d''entrée'!B16))/(2*PI()*'données d''entrée'!B19*'Calculs & résultats'!C7)
        R_cond_tube = math.log(self.de / self.di) / (2 * 3.14159 * self.k_pipe * self.L_tube)
   

        #calcul de la resistance de l'isolant
        R_cond_insulation = math.log(self.de_with_insulation / self.de) / (2 * 3.14159 * self.k_insulation * self.L_tube)

        q_total = q_conv + q_rad
        if R_rad != 0 or R_rad is not None:
            R_total = 1/(1/R_conv + 1/R_rad)
        else:
            R_total = R_conv
        
        # Calculate internal wall temperature
        T_wall_i = self.Tfluid - (q_total * R_conv_i)
        T_wall_e=T_wall_i - (q_total * R_cond_tube)
        Tc=T_wall_e - (q_total * R_cond_insulation)
    

        return q_conv, q_rad, q_total, R_conv, R_rad, R_total, R_conv_i, T_wall_i, T_wall_e, Tc, R_cond_tube

    def calculate_reynolds_number(self):
        # Calculate Reynolds number for the fluid side
        self.v = (self.F_m3h / 3600)*4 / (3.14159 * self.di ** 2)
        Re_fluid = (self.rho * self.v * self.di) / self.mu

        return Re_fluid

    def heat_balance(self, Tc):
        self.Tc = Tc
        self.Tf = (self.Tc + self.Tamb) / 2
        self.rho_air = air_humide.Air_rho_hum(T_db=self.Tf, RH=self.RH, P=101325)
        self.Ra_air, self.Nu, self.h_air = self.calculate_ambient_air_properties()
        self.q_conv, self.q_rad, self.q_total, self.R_conv_e, self.R_rad, self.R_total, self.R_conv_i, self.T_wall_i, self.T_wall_e, T_insulation, self.R_cond_tube = self.calculate_heat_transfer()
        return T_insulation - Tc  # Return the difference for the solver

    def calculate(self):
        Re_fluid = self.calculate_reynolds_number()  # Calculate Reynolds number
        if Re_fluid >= 2300:
            self.regime = 'turbulent'
            self.Nu_fluid = 0.023 * Re_fluid**0.8 * self.Pr_fluid**0.4
        else:
            self.regime = 'laminar'
            self.Nu_fluid = 3.66

        self.h_fluid = self.Nu_fluid * self.k / self.di

        # Use fsolve to find the final value of Tc
        self.Tc = fsolve(self.heat_balance, self.Tc)[0]

        # Recalculate film temperature
        self.Tf = (self.Tc + self.Tamb) / 2
        self.rho_air = air_humide.Air_rho_hum(T_db=self.Tf, RH=self.RH, P=101325)
        self.Ra_air, self.Nu, self.h_air = self.calculate_ambient_air_properties()
        self.q_conv, self.q_rad, self.q_total, self.R_conv_e, self.R_rad, self.R_total, self.R_conv_i, self.T_wall_i, self.T_wall_e, self.Tc, self.R_cond_tube = self.calculate_heat_transfer()

        data = {
            'Fluid': [self.fluid],
            'Regime': [self.regime],
            'T_fluid (°C)': [self.Tfluid],
            'v (m/s)': [self.v],
            'F_m3h (m3/h)': [self.F_m3h],
            'DN': [self.dn * 1000],  # Convert back to mm for display
            'di (m)': [self.di],
            'de (m)': [self.de],
            'L_tube (m)': [self.L_tube],
            'Material': [self.material],
            'Insulation': [self.insulation],
            'Insulation Thickness (m)': [self.insulation_thickness],
            'Emissivity': [self.emissivity],
            'Tamb (°C)': [self.Tamb],
            'Humidity (%)': [self.RH],
            'Tamb (°C)': [self.Tamb],
      
        
            'Flow Regime': [self.regime],
            'Tc (°C)': [self.Tc],
            'Tf (°C)': [self.Tf],
            'Outer Diameter with Insulation (m)': [self.de_with_insulation],
            'Prandtl Number (Pr)': self.Pr_air,
            'External Surface Area (m²)': self.A_e_insulated,
            'Rayleigh Number': self.Ra_air,
            'Nusselt Number': self.Nu,
            'Average Heat Transfer Coefficient (W/m².K)': self.h_air,
            'Convective Heat Transfer (W)': self.q_conv,
            'Radiative Heat Transfer (W)': self.q_rad,
            'Total Heat Transfer (W)': self.q_total,
            'Convective Resistance (K/W)': self.R_conv_e,
            'Radiative Resistance (K/W)': self.R_rad,
            'Equivalent Resistance (K/W)': self.R_total,
            'Internal Surface Area (m²)': [self.A_i],
            'Reynolds Number (Re_fluid)': Re_fluid,  
            'velocity (m/s)': self.v,
            'Prandtl Number (Pr)': self.Pr_fluid,  # Add Prandtl number of the fluid to the data
            'Nusselt Number (self.Nu_fluid)': self.Nu_fluid,  # Add Nusselt number of the fluid to the data
            'Heat Transfer Coefficient (self.h_fluid)': self.h_fluid,  # Add heat transfer coefficient of the fluid to the data
            'Internal Convective Resistance (K/W)': self.R_conv_i,
            'Internal Wall Temperature (°C)': self.T_wall_i,
            'External Wall Temperature (°C)': self.T_wall_e,
            'Insulation Temperature (°C)':  self.Tc,
            'Conductive Resistance of Bare Tube (K/W)': self.R_cond_tube,
        }

        self.df = pd.DataFrame(data).T

# Example usage





import requests
import math

import os
import sys

cwd = os.getcwd()
#print("+++--------------+++++++++=",cwd)
cwd=cwd.replace("\\", "/")
#print("+++++------------++++++++++=",cwd)

#pip install requests[security]
 
sys.path.insert(0, os.path.join( os.path.dirname(__file__), "" , ""))
config_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config.ini')
#print("config path.........................",config_path)

from configparser import ConfigParser


def get_api_key():
    config = ConfigParser()
    config.read(config_path)
    print("change OpenWeatherMap api=xxxxxxxxxxxxxxx if necessary in",config.read(config_path))
    return config['DonneesMeteo']['api']

def get_location():
    config = ConfigParser()
    config.read(config_path)
    #print(config['Location']['Town'])
    return config['Location']['Town'],config['Location']['lat'],config['Location']['lon']

 
def get_weather(api_key, location,lat,lon):
    if math.isnan(float(lat)) ^ math.isnan(float(lon)) :
        url="http://api.openweathermap.org/data/2.5/weather?lat={}&lon={}&appid={}".format(lat,lon, api_key)
       # print("url=",url)
    
    else:
        # url = "http://api.openweathermap.org/data/2.5/weather?q={}&appid={}".format(location, api_key)
         url="http://api.openweathermap.org/data/2.5/weather?lat={}&lon={}&appid={}".format(lat,lon, api_key)
        # print("url_town=",url)
       
    import time

    r = ''
    while r== '':
        try:
            #r = requests.get(url,timeout=(3.05, 27))
            r = requests.get(url,timeout=(5, 5))
          #  print("url reqtest=",url)
          #  print("r=",r)
            break
        except:
            print("Connection refused by the server..")
            print("Let me sleep for 9 seconds")
            time.sleep(9)
            print("Was a nice sleep, now let me continue...")
            continue

   
#    
  #  r = requests.get(url)
    print(r)
    return r.json()
 


#get_api_key()


import pandas as pd

from .apply_heat_exchange_above import apply_heat_exchange_above
from .apply_heat_exchange_below import apply_heat_exchange_below
from .check_additional_heat_exchanges import check_additional_heat_exchanges
from .draw_hen_from_df import draw_hen_from_df
debug_iteration = 1  # Itération pour laquelle le debug est activé

###############################"GHE########################################"""""
def graphical_hen_design(self, plot=False):
    # Initialiser une liste pour les échangeurs installés
    self.heat_exchangers = []
    self.remaining_recoverable_heat = self.heat_recovery  # Track remaining recoverable heat
    self.used_streams = set()  # Track streams that have already been used

    # Initialize remaining stream lists
    self.remain_stream_list_above = self.stream_list_above.copy()
    self.remain_stream_list_below = self.stream_list_below.copy()



    # === Échanges au-dessus du pinch ===
    if not self.combinations_above.empty:
        for i in range(len(self.combinations_above)):
            comb = self.combinations_above.iloc[i]
            one_HS_df = self.remain_stream_list_above[self.remain_stream_list_above['id'] == comb['HS_id']]
            one_CS_df = self.remain_stream_list_above[self.remain_stream_list_above['id'] == comb['CS_id']]

            # Debug: Print the combination being tested

            # print(f"\nEvaluating combination above the pinch: HS_id={comb['HS_id']}, CS_id={comb['CS_id']}")

            if not one_HS_df.empty and not one_CS_df.empty:
                one_HS_df, one_CS_df,self.remain_stream_list_above = apply_heat_exchange_above(one_HS_df, one_CS_df,self.remain_stream_list_above,self.remaining_recoverable_heat,self.used_streams,self.heat_exchangers)

    # === Échanges en-dessous du pinch ===
    if not self.combinations_below.empty:
        for i in range(len(self.combinations_below)):
            comb = self.combinations_below.iloc[i]
            one_HS_df = self.remain_stream_list_below[self.remain_stream_list_below['id'] == comb['HS_id']]
            one_CS_df = self.remain_stream_list_below[self.remain_stream_list_below['id'] == comb['CS_id']]

            # Debug: Print the combination being tested

            # print(f"\nEvaluating combination below the pinch: HS_id={comb['HS_id']}, CS_id={comb['CS_id']}")

            if not one_HS_df.empty and not one_CS_df.empty:
                one_HS_df, one_CS_df,self.remain_stream_list_below = apply_heat_exchange_below(one_HS_df, one_CS_df,self.remain_stream_list_below,self.remaining_recoverable_heat,self.used_streams,self.heat_exchangers)


    # === Vérifier et ajouter les échanges supplémentaires en boucle ===
  

    additional_heat_exchanges_above,self.remain_stream_list_above = check_additional_heat_exchanges(self.remain_stream_list_above)
    # print(additional_heat_exchanges_above)
    additional_heat_exchanges_below,self.remain_stream_list_below = check_additional_heat_exchanges(self.remain_stream_list_below)
    # print(additional_heat_exchanges_below)

    # Ajouter les échanges supplémentaires à la liste des échangeurs
    self.heat_exchangers.extend(additional_heat_exchanges_above)
    self.heat_exchangers.extend(additional_heat_exchanges_below)

    # === Construction finale du DF des échangeurs ===
    self.df_exchangers = pd.DataFrame(self.heat_exchangers)

    if not self.df_exchangers.empty:
        self.df_exchangers = self.df_exchangers[self.df_exchangers['HeatExchanged'] > 0.5]
        self.df_exchangers = self.df_exchangers.sort_values(by="HeatExchanged", ascending=False).reset_index(drop=True)

        # chaleur totale récupérée
        self.total_heat_recovered = self.df_exchangers['HeatExchanged'].sum()
        self.percent_recovered = 100 * self.total_heat_recovered / self.heat_recovery
    else:
        # Aucun échange trouvé
        self.total_heat_recovered = 0.0
        self.percent_recovered = 0.0
    if plot:
        draw_hen_from_df(self.df_exchangers)
    return self.df_exchangers


from .FreshAir import FreshAir
from .air_humide import air_humide
from .Coil import HeatingCoil
from .Coil import CoolingCoil_Sensible
#from .TkinterGUI import TkinterGUI
from .Humidification import Humidifier
from .HeatRecovery import Heat_plate_exchanger
from .GenericAHU import AirRecyclingAHU
from .GenericAHU import AirRecoveryAHU



# Définition d'un type MIME pour la boîte de liste
LISTBOX_MIMETYPE = "application/x-item"

# Codes d'opération pour différents types de nœuds
OP_NODE_INPUT = 10
OP_NODE_INPUT_P_h = 20
OP_NODE_OUTPUT = 30
OP_NODE_MIX = 40
OP_NODE_ADD = 50
OP_NODE_SUB = 60
OP_NODE_MUL = 70
OP_NODE_DIV = 80
OP_NODE_COMP = 90
OP_NODE_COMP_M = 100

OP_NODE_EVAP = 110
OP_NODE_DESU = 120
OP_NODE_COND = 130
OP_NODE_DET = 140
OP_NODE_TURB = 150
OP_NODE_HEXT = 160

OP_NODE_AIR_INPUT = 170
OP_NODE_RANDOM_METEO = 180
OP_NODE_AIR_OUTPUT = 190
OP_NODE_HEATING_COIL = 200
OP_NODE_COOLING_COIL = 210
OP_NODE_HMD = 220
OP_NODE_AIRMIX = 230
OP_NODE_COOLING_COIL_SENSIBLE = 240
OP_NODE_COOLING_COIL_EXPERT = 250

# Dictionnaire des nœuds de calcul
CALC_NODES = {}

# Définition des exceptions pour la configuration
class ConfException(Exception): pass
class InvalidNodeRegistration(ConfException): pass
class OpCodeNotRegistered(ConfException): pass

# Fonction pour enregistrer un nœud maintenant
def register_node_now(op_code, class_reference):
    # Vérification des doublons lors de l'enregistrement d'un nœud
    if op_code in CALC_NODES:
        raise InvalidNodeRegistration("Duplicate node registration of '%s'. There is already %s" % (
            op_code, CALC_NODES[op_code]
        ))
    CALC_NODES[op_code] = class_reference

# Fonction décorateur pour enregistrer un nœud
def register_node(op_code):
    def decorator(original_class):
        register_node_now(op_code, original_class)
        return original_class
    return decorator

# Fonction pour obtenir la classe associée à un code d'opération
def get_class_from_opcode(op_code):
    # Vérification si le code d'opération est enregistré
    if op_code not in CALC_NODES:
        raise OpCodeNotRegistered("OpCode '%d' is not registered" % op_code)
    return CALC_NODES[op_code]

# Importation de tous les nœuds et enregistrement
from PyqtSimulator.nodes import *
